import sys; sys.path.append("C:\\Users\\cac9159\\Repos\\")
import trikit
from trikit import totri, chladder

# Load sample dataset.
raa = trikit.load(dataset="raa")
DATA = raa


# Convert dataset with origin, dev and value to a triangle object:
tri = trikit.totri(DATA, type_="cum")


# Bind reference to latest_diagonal:
latest = tri.latest


# Generate age-to-age factors:
a2a = tri.a2a


# Review all computed ldf averages:
avgs = tri.a2a_avgs


# Visualize triangle development to date as lattice plot:
tri.plot()


# Generate incremental triangle:
itri = tri.as_incr()



res_ = tri.cl(sel="all-weighted", tail=1.0, range_method=None)






# Chain Ladder demonstration =================================================]

# Pass original dataset into top-level chladder function:
cl = chladder(data=DATA, range_method=None)


# In order to produce Chain Ladder reserve estimates, call cl's run method:
clsumm = cl.run(sel="all-weighted", tail=1.0)


# Produce Chain Ladder reserve estimate summary:
clsumm


# View squared traingle:
clsumm.trisqrd.round()

# View projected development along with actual development to date:
clsumm.plot()

# Deomonstration of trikit's range generation functionality ==================]

# Pass original dataset into top-level chladder function, but specify
# "bootstrap" for range_method:
bcl = chladder(data=DATA, range_method="bootstrap")

bclsumm = bcl.run(sims=1000)




"""
<>=<>=<>=<>=<>=<>=<>=<>=<>=<>=<>=<>=<>=<>=<>=<>=<>=<>=<>=<>=<>=<>=<>=<>=<>
|  _                                                                     |
| | |_ _ __(_) | _(_) |_                                                 |
| | __| '__| | |/ / | __|                                                |
| | |_| |  | |   <| | |_                                                 |
|  \__|_|  |_|_|\_\_|\__|                                                |
|                                                                        |
| Actuarial Reserving Methods in Python                                  |
|_______________________________________________________________________ |
|                                                                        |
| Created by      => James D Triveri <<<james.triveri@gmail.com>>>       |
| License         => 3-Clause BSD (See LICENSE in top-level directory    |
| Repository Link => https://github.com/jtrive84/trikit                  |                                                                  |
<>=<>=<>=<>=<>=<>=<>=<>=<>=<>=<>=<>=<>=<>=<>=<>=<>=<>=<>=<>=<>=<>=<>=<>=<>

Outstanding Tasks =>

[*] Find out how to suppress scientific notation without distorting the
     appearance of NaN's in DataFrames

[*] Implement BFChainLadder

[*] Implement MCMCChainLadder

[*] Add neg_handler #2 logic in _BootstrapChainLadder _bs_samples

[*] Add examples to *ChainLadder docstrings.

[*] Create trikit matplotlib stylesheet: https://matplotlib.org/users/style_sheets.html

[*] Accept additional styling arguments in triangler and chanladder plot method.

[*] Fix _BaseChainLadder's run method to mirror BootstrapChainLadder's run method.

[*] Update documentation for _ChainLadderResult

[*] Fix this error:
# tri2 = DataFrame
trikit._IncrTriangle(tri2)
# Throws error:
# Traceback (most recent call last):
#   File "C:\Python37\lib\site-packages\IPython\core\interactiveshell.py", line 3296, in run_code
#     exec(code_obj, self.user_global_ns, self.user_ns)
#   File "<ipython-input-75-5ef5b2b85c97>", line 1, in <module>
#     trikit._IncrTriangle(tri2)
#   File "G:\Repos\trikit\triangle.py", line 75, in __init__
#     for i in tri:
# UnboundLocalError: local variable 'tri' referenced before assignment

https://sphinxcontrib-napoleon.readthedocs.io/en/latest/index.html
"""
import os
import sys
import collections
import datetime
import os.path
import warnings
import numpy as np
import pandas as pd
import scipy
# import matplotlib as mpl
# import matplotlib.pyplot as plt
# import seaborn as sns
from .triangle import _IncrTriangle, _CumTriangle
from .chainladder import _BaseChainLadder
from .chainladder.mack import _MackChainLadder
from .chainladder.bootstrap import _BootstrapChainLadder
from .datasets import dataref
from .utils import (
    _load, _get_datasets, _get_lrdb_lobs, _get_lrdb_groups, _get_lrdb_specs,
    totri, _cumtoincr, _incrtocum, _tritotbl, _get_datasets #,plot_devp
    )

# Initialize `_load` closure using dataref mapping.
lrdb_path = dataref["lrdb"]
load = _load(dataref=dataref)
get_datasets = _get_datasets(dataref=dataref)
get_lrdb_lobs = _get_lrdb_lobs(lrdb_path=lrdb_path)
get_lrdb_groups = _get_lrdb_groups(lrdb_path=lrdb_path)
get_lrdb_specs = _get_lrdb_specs(lrdb_path=lrdb_path)

pd.options.mode.chained_assignment = None # 'warn'
pd.set_option('display.max_columns', 1000)
pd.set_option('display.width', 500)
np.set_printoptions(
    edgeitems=3, linewidth=200, suppress=True, nanstr='NaN',
    infstr='Inf', precision=5
    )

from .__pkginfo__ import version as __version__
from .__pkginfo__ import name



def chladder(data, origin=None, dev=None, value=None, trifmt=None,
             datafmt="incremental", range_method=None):
    """
    Return *ChainLadder initializer with optional reserve range method
    specification.

    Parameters
    ----------
    data: pd.DataFrame
        The dataset to be coerced into a *Triangle instance. ``data`` can be
        tabular loss data or a pandas DataFrame formatted as a triangle but
        not typed as such. In the latter case, ``trifmt`` must be specified to
        indicate the representation of ``data`` (either "cumulative" or
        "incremental").

    origin: str
        The field in ``data`` representing the origin year. When
        ``trifmt`` is not None, ``origin`` is ignored. Defaults to None.

    dev: str
        The field in ``data`` representing the development period. When
        ``trifmt`` is not None, ``dev`` is ignored. Defaults to None.

    value: str
        The field in ``data`` representing loss amounts. When ``trifmt`` is
        not None, ``value`` is ignored. Defaults to None.

    trifmt: str
        One of "cumulative", "incremental" or None (None by default).
        ``trifmt`` should only be set to something other than None if ``data``
        is a DataFrame formatted as a loss triangle, but hasn't yet been
        converted to a ``CumTriangle`` or ``IncrTriangle`` instance. When
        ``datafmt`` is not None, ``trifmt`` is ignored.

    datafmt: str
        When ``data`` is in tabular form, ``datafmt`` indicates whether the
        records comprising ``data`` represent cumulative or incremental
        losses.  When ``trifmt`` is not None, ``datafmt`` is ignored. Default
        value is "incremental".

    range_method: str
        One of "mack", "bootstrap" or None. Additional methods will
        be available in future releases. Defaults to None.

    Returns
    -------
    chainladder.*ChainLadder instance.
        An instance of chainladder.*ChainLadder, optionally with ranges
        quantifying reserve variability.

    Examples
    --------
    Here we demonstrate how to produce Chain Ladder estimates for the
    ``raa`` sample dataset. Note that what is returned by ``chladder``
    is callable: In order to obtain the Chain Ladder summary, the
    object returned by ``chladder`` must be called, optionally
    specifying ``sel`` (defaults to "all-weighted") and ``tail`` (defaults
    to 1.0). Implemented this way, the Chain Ladder ultimates resulting
    from any number of `loss development factor - tail factor` combinations
    can be specified for the same triangle and compared, in preference
    to instantiating a new ``_BaseChainLadder`` instance each time ``sel``
    or ``tail`` should be varied.

    >>> import trikit
    >>> data = trikit.load("ta83")
    >>> cl_init = trikit.chladder(data=data)
    >>> cl_init(sel="all-weighted", tail=1.005)
       origin maturity    latest      cldf       ultimate       reserve
    0    1981       10   18834.0  1.005000   18928.170000     94.170000
    1    1982        9   16704.0  1.014263   16942.243687    238.243687
    2    1983        8   23466.0  1.031441   24203.787778    737.787778
    3    1984        7   27067.0  1.065750   28846.657874   1779.657874
    4    1985        6   26180.0  1.110442   29071.370025   2891.370025
    5    1986        5   15852.0  1.236349   19598.608700   3746.608700
    6    1987        4   12314.0  1.448599   17838.049103   5524.049103
    7    1988        3   13112.0  1.841007   24139.288472  11027.288472
    8    1989        2    5395.0  2.988917   16125.209021  10730.209021
    9    1990        1    2063.0  8.964835   18494.454742  16431.454742
    10  total           160987.0       NaN  214187.839403  53200.839403

    """
    tri = totri(data=data, type_="cumulative", origin=origin, dev=dev,
                value=value, trifmt=trifmt, datafmt=datafmt)

    if range_method is None:
        # Instantiate BaseChainLadder instance.
        cl_init = _BaseChainLadder(cumtri=tri)

    elif range_method and range_method.lower().startswith("mack"):
        # Instantiate _MackChainLadder instance.
        cl_init = _MackChainLadder(cumtri=tri)

    elif range_method and range_method.lower().startswith("boot"):
        # Instantiate _BootstrapChainLadder instance.
        cl_init = _BootstrapChainLadder(cumtri=tri)

    else:
        raise NotImplementedError(
            "Specified range_method does not exist: `{}`".format(range_method)
            )

    return(cl_init)




"""
Various trikit utilities. Contains convenience functions in support of
the CAS Loss Reserving Database, as well as the intended primary interface
with which to transform tabular data to Triangle objects (``totri``).
"""
import sys
import os.path
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from numpy.random import RandomState
from .triangle import _IncrTriangle, _CumTriangle




def _load(dataref):
    """
    trikit's sample dataset loading utility.
    """
    def func(dataset, loss_type="paid", lob=None, grcode=None,
             grname=None, upper_left_ind=True, lower_right_ind=False,
             allcols=False, action=None, random_state=None):
        """
        Load the specified dataset, returning a DataFrame of incremental
        losses. If ``dataset`` ="lrdb", additional keyword arguments are used
        to subset the CAS Loss Reserving Database to the records of interest.
        Within the Loss Reserving Database, "loss_key" and "grname" uniquely
        partition losses into 100 record blocks if ``lower_right_ind`` =True,
        otherwise losses are partitioned into 55 record blocks. All available
        combinations of "loss_key" and "grcode" (referred to as "specs")
        can be obtained by calling ``get_lrdb_specs``.
        If ``dataset`` is something other than "lrdb", then only the name of
        the target dataset as a string is required.

        Parameters
        ----------
        dataset: str
            Specifies which sample dataset to load. The complete set of sample
            datasets can be obtained by calling ``get_datasets``.

        lob: str
            One of "WC", "COM_AUTO", "MED_MAL", "OTHR_LIAB", "PP_AUTO" or
            "PROD_LIAB". When ``dataset`` ="lrdb", specifies which losses to
            target. The complete mapping of available lobs can be obtained by
            calling ``get_lrdb_lobs``. Applies only when ``dataset`` ="lrdb",
            otherwise parameter is ignored.

        grcode: str
            NAIC company code including insurer groups and single insurers.
            The complete mapping of available grcodes can be obtained by
            calling ``get_lrdb_groups``. Applies only when
            ``dataset`` ="lrdb", otherwise parameter is ignored.

        grname: str
            NAIC company name (including insurer groups and single insurers).
            The complete mapping of available grcodes can be obtained by
            calling ``get_lrdb_groups``. Applies only when
            ``dataset`` ="lrdb", otherwise parameter is ignored.

        loss_type: str
            Specifies which loss data to load. Can be one of "paid",
            "incurred" or "bulk". Defaults to "paid". Applies only when
            ``dataset`` ="lrdb", otherwise parameter is ignored.

        upper_left_ind: bool
            If True, the upper-left portion of the triangle will be returned.
            The upper-left portion of the triangle typically consists of
            actual loss experience. Defaults to True. Applies only when
            ``dataset`` ="lrdb", otherwise parameter is ignored.

        lower_right_ind: bool
            If True, the lower-right portion of the triangle will be returned.
            The CAS Loss Reserve Database includes 10 development lags for
            each loss year, the intention being that the data comprising the
            lower-right portion of the triangle can be used for model
            validation purposes. Defaults to False. Applies only when
            ``dataset`` ="lrdb", otherwise parameter is ignored.

        allcols: bool
            If True, the returned DataFrame contains all columns that comprise
            the CAS Loss Reserving Database. Defaults to False. When
            ``allcols`` =False, only the columns required to convert the
            dataset to an ``IncrTriangle`` or ``CumTriangle`` instance, namely
            "origin", "dev" and "value", with the field specified by ``loss``
            renamed to "value" will be returned. Note that if
            ``action`` ="aggregate", ``allcols`` is bound to False,
            regardless of how it was originally specified in the function
            call.

        action: str
            Action to perform if the specified subsetting parameters do not
            reduce to a single "loss_key"-"grcode". If more than one
            combination of "loss_key" and "grcode" remains after applying all
            subsetting specifications, the remaining records can either be
            (1) aggregated (``action`` ="aggregate"), (2) a single
            "loss_key"-"grcode" combination can be selected at random and the
            associated records returned (``action`` ="random") or (3) the
            remaining records can be returned as-is without additional
            processing (``action`` =None). Defaults to None. Note that
            ``action``="aggregate" implicitly sets ``allcols`` to False,
            regardless of how the parameter may have been set initially.
            Applies only when ``dataset`` ="lrdb", otherwise parameter is
            ignored.

        random_state: RandomState/int
            If int, random_state is the seed used by the random number
            generator; If RandomState instance, random_state is the random
            number generator; If None, the random number generator is the
            RandomState instance used by np.random. Applies only when
            ``dataset ="lrdb"``, otherwise parameter is ignored.

        Returns
        -------
        pd.DataFrame
            Sample dataset or subset of the CAS Loss Reserving Database.
        """
        try:
            dataset  = dataset.lower()
            datapath = dataref[dataset]
            dat_init = pd.read_csv(datapath, delimiter=",")

        except KeyError:
            print("Specified dataset does not exist: `{}`".format(dataset))

        # Additional filtering/subsetting if dataset="lrdb".
        if dataset=="lrdb":
            if   loss_type.lower().startswith("i"): loss_field = "incrd_loss"
            elif loss_type.lower().startswith("p"): loss_field = "paid_loss"
            elif loss_type.lower().startswith("b") : loss_field = "bulk_loss"

            basecols = ["loss_key", "grcode", "origin", "dev", loss_field]
            lower_right_spec = 1 if lower_right_ind==True else 0
            upper_left_spec  = 1 if upper_left_ind==True  else 0
            loss_key_spec, loss_type_spec = lob, loss_type
            grname_spec, grcode_spec = grname, grcode

            if all([lower_right_spec, upper_left_spec]):
                dat_init = dat_init[
                    (dat_init["lower_right_ind"]==lower_right_spec) |
                    (dat_init["upper_left_ind"]==upper_left_spec)
                    ]
            elif upper_left_spec==1 and lower_right_spec==0:
                dat_init = dat_init[
                    dat_init["upper_left_ind"]==upper_left_spec
                    ]
            elif upper_left_spec==0 and lower_right_spec==1:
                dat_init = dat_init[
                    dat_init["lower_right_ind"]==lower_right_spec
                    ]
            else:
                raise ValueError(
                    "At least one of upper_left_ind or lower_right_ind must be 1."
                    )

            if loss_key_spec is not None:
                dat_init = dat_init[dat_init["loss_key"]==loss_key_spec]

            if grname_spec is not None:
                dat_init = dat_init[dat_init["grname"]==grname_spec]

            if grcode_spec is not None:
                dat_init = dat_init[dat_init["grcode"]==grcode_spec]

            # Check whether provided filter specs filter down to a single
            # loss_key-grname/grcode combination. If more than one spec
            # remains, look to `action` parameter.
            fields = ["loss_key", "grcode"]
            remaining_specs = dat_init[fields].drop_duplicates().reset_index(drop=True)

            if remaining_specs.shape[0] > 1:

                if action and action.lower().startswith("agg"):
                    # Aggregate remaining records over `origin` and `dev`.
                    # Requires dropping `loss_key` and `grcode`.
                    dat = dat_init[["origin", "dev", loss_field]].groupby(["origin", "dev"],
                        as_index=False).sum().reset_index(drop=True)

                elif action and action.lower().startswith("rand"):
                    # Check random_state and initialize random number generator.
                    if random_state is not None:
                        if isinstance(random_state, int):
                            prng = RandomState(random_state)
                        elif isinstance(random_state, RandomState):
                            prng = random_state
                    else:
                        prng = RandomState()

                    # Randomly select record from remaining_specs.
                    keepspec = remaining_specs.loc[prng.choice(remaining_specs.index)]
                    keep_loss_key, keep_grcode = keepspec.loss_key, keepspec.grcode

                    # Filter dat_init using keep_*-prefixed fields.
                    dat = dat_init[
                        (dat_init.loss_key==keep_loss_key) & (dat_init.grcode==keep_grcode)
                        ]

                    dat = dat[basecols]

                else: # Return dataset "as-is", containing > 1 specs.
                    dat = dat_init[basecols]
            else:
                dat = dat_init[basecols]

        else: # Specified dataset is not "lrdb".
            loss_field = "value"
            basecols = ["origin", "dev", "value"]
            dat      = dat_init

        if not allcols:
            dat = dat.rename({loss_field:"value"}, axis=1)
        return(dat.reset_index(drop=True))
    return(func)





# Loss Reserving Database utility functions ==================================]
def _get_datasets(dataref:dict):
    def func():
        """
        Generate a list containing the names of available sample datasets.

        Returns
        -------
        list
            Names of available sample datasets.
        """
        return(list(dataref.keys()))
    return(func)



def _get_lrdb_lobs(lrdb_path:str):
    """
    Return the unique "loss_key" entries present in the CAS Loss
    Reserving Database (lrdb).
    """
    lrdb = pd.read_csv(lrdb_path, sep=",")
    lrdb = lrdb["loss_key"].unique()
    def func():
        """
        Return unique "loss_key" entries from the CAS Loss Reserving
        Database.

        Returns
        -------
        list
            List containing unique "loss_key" entries.
        """
        return(lrdb.tolist())
    return(func)



def _get_lrdb_groups(lrdb_path:str):
    """
    Return "grcode"-"grname" mapping present in the CAS Loss
    Reserving Database (lrdb).
    """
    fields = ["grcode", "grname"]
    lrdb   = pd.read_csv(lrdb_path, sep=",")
    lrdb   = lrdb[fields].drop_duplicates().reset_index(drop=True)
    def func(returnas=pd.DataFrame):
        """
        Return lrdb groups mapping in accordance with object class
        specified in ``returnas`` argument.

        Parameters
        ----------
        returnas: type
            The object class used to represent the lrdb "grcode"-"grname"
            mapping. Valid types include pd.DataFrame, list or dict.
            Any other object passed to ``returnas`` raises TypeError. Default
            type is pd.DataFrame.

        Returns
        -------
        list, dict, pd.DataFrame
            Unique combinations of "grcode"-"grname" as present in the CAS
            Loss Reserving Database.
        """
        if returnas==pd.DataFrame:
            groups = lrdb
        elif returnas==list:
            groups_init = {j:i for i, j in set(zip(lrdb.grname, lrdb.grcode))}
            groups = list(groups_init.items())
        elif returnas==dict:
            groups = {j:i for i, j in set(zip(lrdb.grname, lrdb.grcode))}
        else:
            raise TypeError("Invalid returnas object class: `{}`".format(returnas))
        return(groups)
    return(func)



def _get_lrdb_specs(lrdb_path:str):
    """
    Return a DataFrame containing the unique combinations of "loss_key",
    "grname" and "grcode" from the CAS Loss Reserving Database (lrdb).
    """
    fields = ["loss_key", "grcode", "grname"]
    lrdb = pd.read_csv(lrdb_path, sep=",")
    lrdb = lrdb[fields].drop_duplicates().reset_index(drop=True)
    lrdb = lrdb.sort_values(by=["loss_key","grcode"])
    def func():
        """
        Return a DataFrame containing the unique combinations of "loss_key",
        "grname" and "grcode" from the CAS Loss Reserving Database (lrdb).

        Returns
        -------
        pd.DataFrame
            Unique combinations of "loss_key", "grname" and "grcode"
            from the CAS Loss Reserving Database.
        """
        return(lrdb)
    return(func)




def totri(data=None, type_="cumulative", origin="origin", dev="dev",
          value="value", trifmt=None, datafmt="incremental"):
    """
    Transform ``data`` to a triangle object. ``type_`` can be one of
    "incremental" or "cumulative". If ``trifmt==None``, it is assumed
    ``data`` is tabular loss data (e.g., pd.DataFrame) containing at
    minimum the fields "origin", "dev" and "value". If ``trifmt`` is set
    to "incremental" or "cumulative", ``data`` is assumed to be formatted
    as a loss triangle, but hasn't yet been transformed into a ``_CumTriangle``
    or _IncrTriangle instance. Returns either a ``_CumTriangle`` or
    ``_IncrTriangle`` instance, depending on the value passed to ``type_``.

    Parameters
    ----------
    data: pd.DataFrame
        The dataset to be coerced into a *Triangle instance. ``data`` can be
        tabular loss data or a pandas DataFrame formatted as a triangle but
        not typed as such. In the latter case, ``trifmt`` must be specified to
        indicate the representation of ``data`` (either "cumulative" or
        "incremental").

    type_: str
        Either "cumulative" or "incremental". Alternatively, ``type_`` can be
        specified as "i" for "incremental" or "c" for "cumulative". Dictates
        the type of triangle returned. Default value is "cumulative".

    trifmt: str
        One of "cumulative", "incremental" or None (None by default).
        ``trifmt`` should only be set to something other than None if ``data``
        is a DataFrame formatted as a loss triangle, but hasn't yet been
        converted to a ``CumTriangle`` or ``IncrTriangle`` instance. When
        ``datafmt`` is not None, ``trifmt`` is ignored.

    datafmt: str
        When ``data`` is in tabular form, ``datafmt`` indicates whether the
        records comprising ``data`` represent cumulative or incremental
        losses.  When ``trifmt`` is not None, ``datafmt`` is ignored. Default
        value is "incremental".

    origin: str
        The field in ``data`` representing the origin year. When
        ``trifmt`` is not None, ``origin`` is ignored. Defaults to None.

    dev: str
        The field in ``data`` representing the development period. When
        ``trifmt`` is not None, ``dev`` is ignored. Defaults to None.

    value: str
        The field in ``data`` representing loss amounts. When ``trifmt`` is
        not None, ``value`` is ignored. Defaults to None.

    Returns
    -------
    {trikit.triangle._IncrTriangle, trikit.triangle._CumTriangle}
    """
    if trifmt:
        if trifmt.lower().startswith("i"):
            # `data` is in incremental triangle format (but not typed as such).
            data2 = _tritotbl(data)
        elif trifmt.lower().startswith("c"):
            # `data` is in cumulative triangle format (but not typed as such).
            data2 = _tritotbl(_cumtoincr(data))
        else:
            raise NameError("Invalid type_ argument: `{}`.".format(type_))

    else:
        if datafmt.lower().startswith("c"): # `data` is in cumulative tabular format.
            data2 = data.copy(deep=True)
            data2["incr"] = data2.groupby([origin])[value].diff(periods=1)
            data2["incr"] = np.where(np.isnan(data2["incr"]), data2[value], data2["incr"])
            data2 = data2.drop(value, axis=1).rename({"incr":value}, axis=1)

        else: # `data` is in incremental tabular format.
            data2 = data

    if type_.lower().startswith("i"):
        tri = _IncrTriangle(data=data2, origin=origin, dev=dev, value=value)
    elif type_.lower().startswith("c"):
        tri = _CumTriangle(data=data2, origin=origin, dev=dev, value=value)
    return(tri)



def _cumtoincr(cumtri):
        """
        Convert a cumulative triangle to incremental representation.
        Not intended for public use.

        Parameters
        ----------
        cumtri: cumulative.CumTriangle
            An instance of cumulative.CumTriangle.

        Returns
        -------
        pd.DataFrame
        """
        tri = pd.DataFrame().reindex_like(cumtri)
        tri.iloc[:,0] = cumtri.iloc[:,0]
        for i in enumerate(tri.columns[1:], start=1):
            indx, col = i[0], i[1]
            tri.iloc[:,indx] = cumtri.iloc[:,indx] - cumtri.iloc[:,(indx - 1)]
        return(tri)



def _incrtocum(incrtri):
    """
    Convert incremental triangle to cumulative representation. Not
    intended for public use.

    Parameters
    ----------
    incrtri: incremental.IncrTriangle
        An instance of incremental.IncrTriangle.

    Returns
    -------
    pd.DataFrame
    """
    return(incrtri.cumsum(axis=1))



def _tritotbl(tri):
    """
    Convert data formatted as triangle to tabular representation. Fields
    will be identified as "origin", "dev" and "value". Not intended for
    public use.

    Parameters
    ----------
    tri: pd.DataFrame
        The Triangle object or DataFrame to transform to tabular data.

    Returns
    -------
    pd.DataFrame
    """
    tri = tri.reset_index(drop=False).rename({"index":"origin"}, axis=1)
    df = pd.melt(tri, id_vars=["origin"], var_name="dev", value_name="value")
    df = df[~np.isnan(df["value"])].astype({"origin":np.int_, "dev":np.int_, "value":np.float_})
    return(df.sort_values(by=["origin", "dev"]).reset_index(drop=True))



import sys; sys.path.append("E:\\Repos\\")
import unittest
import trikit
import pandas as pd
import numpy as np
import os
import os.path
import logging
import timeit


from ..triangle import incremental, cumulative
from ..utils import _load, totri, _cumtoincr, _incrtocum, _tritotbl
from ..datasets import dataref

load = _load(dataref)


raa  = trikit.load(dataset="raa")
ta83 = trikit.load(dataset="ta83")
auto = trikit.load(dataset="lrdb", loss_key="COM_AUTO", grcode=32743)
nstd = trikit.load(dataset="nonstd")

ctri = totri(data=raa, datatype="incremental", tritype="cumulative")


# check that s.split fails when the separator is not a string
# with self.assertRaises(TypeError):
    # s.split(2)



class TriangleFunctionsTestCase(unittest.TestCase):







class CumTriangleTestCase(unittest.TestCase):
    def setUp(self):
        self.ctri = totri(raa, type_="cumulative")

    def test_a2adim(self):
        self.assertEqual(
            self.ctri.shape[0]-1, self.ctri.a2a.shape[0],
            "Age-toage Factors not properly computed."
            )

    def test_latest_by_origin(self):
        self.assertEqual(
            self.ctri.shape[0]-1, self.ctri.a2a.shape[0],
            "Age-toage Factors not properly computed."
            )





#
class ChainLadderTestCase(unittest.TestCase):
    def setUp(self):
        self.cl = trikit.BaseChainLadder(data=raa, sel="all-weighted", tail=1.0)


    def test_trisqrd(self):
        """
        Verify that self.tri and self.trisqrd are the same in
        upper left.
        """
        self.assertEqual(
            (self.cl.trisqrd-self.cl.tri).sum().sum(),0.0,
            "trisqrd not correctly implemented."
            )


    def test_ultimates(self):
        """
        Verify that ultimates matches the last column of trisqrd.
        """
        atults = self.cl.ultimates
        tsults = self.cl.trisqrd.loc[:,self.cl.trisqrd.columns[-1]]
        self.assertEqual(
            atults, tsults, "Difference in ultimate results"
            )

    def test_reserves(self):
        """
        Test value consistency.
        """
        assertEqual(
            (self.cl.ultimates-self.cl.latest_by_origin-self.cl.reserves).sum(),
            0, "Inconsistency in ults, latest and reserves."
            )






# Bootstrap Unit Tests ########################################################

class BootChainLadderTestCase(unittest.TestCase):
    def setUp(self):
        self.bcl = trikit.BootChainLadder(data=raa, sel="all-weighted", tail=1.0)


    def test_trisqrd(self):
        """
        Verify that self.tri and self.trisqrd are the same in
        upper left.
        """
        self.assertEqual(
            (self.cl.trisqrd-self.cl.tri).sum().sum(),0.0,
            "trisqrd not correctly implemented."
            )


    def test_ultimates(self):
        """
        Verify that ultimates matches the last column of trisqrd.
        """
        atults = self.cl.ultimates
        tsults = self.cl.trisqrd.loc[:,self.cl.trisqrd.columns[-1]]
        self.assertEqual(
            atults, tsults, "Difference in ultimate results"
            )

    def test_reserves(self):
        """
        Test value consistency.
        """
        assertEqual(
            (self.cl.ultimates-self.cl.latest_by_origin-self.cl.reserves).sum(),
            0, "Inconsistency in ults, latest and reserves."
            )



"""
This module contains the class definitions for ``_BaseChainLadder``.
Users should avoid calling any ``*ChainLadder`` instances directly; rather the
dataset and triangle arguments should be passed to ``chladder``, which will
return the initialized ChainLadder instance, from which estimates of outstanding
liabilities and optionally ranges can be obtained.
"""
import pandas as pd
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
import seaborn as sns


class _BaseChainLadder:
    """
    From the Casualty Actuarial Society's "Estimating Unpaid Claims Using
    Basic Techniques" Version 3 (Friedland, Jacqueline - 2010), the
    development method ('Chain Ladder') consists of seven basic steps:

    1. Compile claims data in a development triangle
    2. Calculate age-to-age factors
    3. Calculate averages of the age-to-age factors
    4. Select claim development factors
    5. Select tail factor
    6. Calculate cumulative claims
    7. Project ultimate claims
    """
    def __init__(self, cumtri):
        """
        Generate point estimates for outstanding claim liabilities at
        ultimate for each origin year and in aggregate. The
        ``_BaseChainLadder`` class exposes no functionality to estimate
        variability around the point estimates at ultimate.

        Parameters
        ----------
        cumtri: triangle._CumTriangle
            A cumulative.CumTriangle instance.
        """
        self.tri = cumtri



    def run(self, sel="all-weighted", tail=1.0):
        """
        Compile a summary of ultimate and reserve estimates resulting from
        the application of the development technique over ``self.tri``.
        Generated DataFrame is comprised of origin year, maturity of origin
        year, loss amount at latest evaluation, cumulative loss development
        factors, projected ultimates and the reserve estimate, by origin
        year and in aggregate.

        Parameters
        ----------
        sel: str
            The ldf average to select from ``triangle._CumTriangle.a2a_avgs``.
            Defaults to "all-weighted".

        tail: float
            Tail factor. Defaults to 1.0.

        Returns
        -------
        trikit.chainladder._ChainLadderResult
        """
        # sel = "all-weighted"
        # tail = 1.0
        # ldfs_ = cl._ldfs(sel=sel, tail=tail)
        # cldfs_ = cl._cldfs(ldfs=ldfs_)
        # ultimates_ = cl._ultimates(cldfs=cldfs_)
        # reserves_ = cl._reserves(ultimates=ultimates_)
        # maturity_ = tri1.maturity.astype(np.str)
        # latest_ = tri1.latest_by_origin

        ldfs_ = self._ldfs(sel=sel, tail=tail)
        cldfs_ = self._cldfs(ldfs=ldfs_)
        ultimates_ = self._ultimates(cldfs=cldfs_)
        reserves_ = self._reserves(ultimates=ultimates_)
        maturity_ = self.tri.maturity.astype(np.str)
        latest_ = self.tri.latest_by_origin
        dfcldfs = cldfs_.to_frame().reset_index(drop=False)
        dfcldfs = dfcldfs.rename({"index":"maturity"}, axis=1)
        dfcldfs["maturity"] = dfcldfs["maturity"].astype(np.str)
        dfmat = maturity_.to_frame().reset_index(drop=False)
        dfmat = dfmat.rename({"index":"origin"}, axis=1)
        dfcldfs = pd.merge(dfmat, dfcldfs, on="maturity", how="inner")
        dfcldfs = dfcldfs.set_index("origin").drop("maturity", axis=1)
        dfcldfs.index.name = None

        # Use join instead of merge to preserve origin year indicies.

        # Compile Chain Ladder point estimate summary.
        # dfmatur_ = maturity_.to_frame().reset_index(drop=False).rename({"index":"origin"}, axis=1)
        # dfcldfs_ = cldfs_.to_frame().reset_index(drop=False).rename({"index":"maturity"}, axis=1)
        # dfcldfs_["maturity"] = dfcldfs_["maturity"].astype(np.str)
        # dfsumm = dfmatur_.merge(dfcldfs_, on=["maturity"], how="left").set_index("origin")
        # dfsumm.index.name = None
        # dflatest_ = latest_.to_frame().rename({"latest_by_origin":"latest"}, axis=1)
        # dfultimates_ = ultimates_.to_frame()
        # dfreserves_ = reserves_.to_frame()
        # dfsumm = functools.reduce(
        #     lambda df1, df2: df1.join(df2),
        #     (dflatest_, dfultimates_, dfreserves_), dfsumm
        #     )
        #

        dfsumm = maturity_.to_frame().join(latest_)
        dfsumm = dfsumm.rename({"latest_by_origin":"latest"}, axis=1)
        dfsumm = dfsumm.join(dfcldfs).join(ultimates_).join(reserves_)
        dfsumm.loc["total"] = dfsumm.sum()

        # Set to NaN fields that shouldn't be aggregated.
        dfsumm.loc["total", "maturity"] = ""
        dfsumm.loc["total", "cldf"] = np.NaN
        dfsumm = dfsumm.reset_index().rename({"index":"origin"}, axis=1)
        kwds = {"sel":sel, "tail":tail}

        # Initialize and return _ChainLadderResult instance.
        clresult_ = _ChainLadderResult(
            summary=dfsumm, tri=self.tri, ldfs=ldfs_, cldfs=cldfs_,
            latest=latest_, maturity=maturity_, ultimates=ultimates_,
            reserves=reserves_, **kwds)
        return(clresult_)



    def _ldfs(self, sel="all-weighted", tail=1.0):
        """
        Lookup loss development factors corresponding to ``sel``.

        Parameters
        ----------
        sel: str
            The ldf average to select from ``triangle._CumTriangle.a2a_avgs``.
            Defaults to "all-weighted".

        tail: float
            Tail factor. Defaults to 1.0.

        Returns
        -------
        pd.Series
        """
        try:
            ldfs_ = self.tri.a2a_avgs.loc[sel]
            tindx_ = ldfs_.index.max() + 1
            ldfs_ = ldfs_.append(pd.Series(data=[tail], index=[tindx_]))
        except KeyError:
                print("Invalid age-to-age selection: `{}`".format(sel))
        ldfs_ =pd.Series(data=ldfs_, index=ldfs_.index, dtype=np.float_, name="ldf")
        return(ldfs_.sort_index())


    def _cldfs(self, ldfs):
        """
        Calculate cumulative loss development factors factors by successive
        multiplication beginning with the tail factor and the oldest
        age-to-age factor. The cumulative claim development factor projects
        the total growth over the remaining valuations. Cumulative claim
        development factors are also known as "Age-to-Ultimate Factors"
        or "Claim Development Factors to Ultimate".

        Parameters
        ----------
        ldfs: pd.Series
            Selected ldfs, typically the output of calling ``self._ldfs``.

        Returns
        -------
        pd.Series
        """
        cldfs_indx = ldfs.index.values
        cldfs_ = np.cumprod(ldfs.values[::-1])[::-1]
        cldfs_ = pd.Series(data=cldfs_, index=ldfs.index.values, name="cldf")
        return(cldfs_.astype(np.float_).sort_index())


    def _ultimates(self, cldfs):
        """
        Ultimate claims are equal to the product of the latest valuation of
        losses (the amount along latest diagonal of any ``_CumTriangle``
        instance) and the appropriate cldf/age-to-ultimate factor. We
        determine the appropriate age-to-ultimate factor based on the age
        of each origin year relative to the evaluation date.

        Parameters
        ----------
        cldfs: pd.Series
            Cumulative loss development factors, conventionally obtained
            via _BaseChainLadder's ``_cldfs`` method.

        Returns
        -------
        pd.Series
        """
        ultimates_ = pd.Series(
            data=self.tri.latest_by_origin.values * cldfs.values[::-1],
            index=self.tri.index, name="ultimate"
            )
        return(ultimates_.astype(np.float_).sort_index())


    def _reserves(self, ultimates):
        """
        Return IBNR/unpaid claim estimates. Represents the difference
        between ultimate loss projections for each origin year and the latest
        cumulative loss amount.
        Note that because outstanding claim liabilities are different based
        on the type of losses represented in the triangle ("ibnr" if losses
        are reported/incurred, "unpaid" if data represents paid losses), we
        use the generic term "reserves" to represent either. In any case,
        what is referred to as "reserves" will always equate to the
        difference between the ultimate loss projections and cumulative
        loss amounts at the latest evaluation period for each origin year.

        Parameters
        ----------
        ultimates: pd.Series
            Estimated ultimate losses, conventionally obtained from
            _BaseChainLadder's ``_ultimates`` method.

        Returns
        -------
        pd.Series
        """
        reserves_ = pd.Series(
            data=ultimates - self.tri.latest_by_origin,
            index=self.tri.index, name='reserve')
        return(reserves_.astype(np.float_).sort_index())










class _ChainLadderResult:
    """
    Curated output resulting from ``_BaseChainLadder``'s ``run`` method.
    """
    def __init__(self, summary, tri, ldfs, cldfs, latest, maturity,
                 ultimates, reserves, **kwargs):
        """
        Container object for ``_BaseChainLadder`` output.

        Parameters
        ----------
        summary: pd.DataFrame
            Chain Ladder summary compilation.

        tri: trikit.triangle._CumTriangle
            A cumulative triangle instance.

        ldfs: pd.Series
            Loss development factors.

        cldfs: pd.Series
            Cumulative loss development factors.

        latest: pd.Series
            Latest loss amounts by origin.

        maturity: pd.Series
            Represents ther maturity of each origin relative to development
            period.

        ultimates: pd.Series
            Represents Chain Ladder ultimate projections.

        reserves: pd.Series
            Represents the projected reserve amount. For each origin period,
            this equates to the ultimate loss projection minus the latest
            loss amount for the origin period (reserve = ultimate - latest).

        kwargs: dict
            Additional keyword arguments passed into ``_BaseChainLadder``'s
            ``run`` method.
        """
        self.ultimates = ultimates
        self.reserves = reserves
        self.summary = summary
        self.cldfs = cldfs
        self.ldfs = ldfs
        self.tri = tri

        if kwargs is not None:
            for key_ in kwargs:
                setattr(self, key_, kwargs[key_])

        # Properties.
        self._trisqrd = None

        self.summspecs = {"ultimate":"{:.0f}".format, "reserve":"{:.0f}".format,
                          "latest":"{:.0f}".format, "cldf":"{:.5f}".format,}


    @property
    def trisqrd(self):
        """
        Project claims growth for each future development period. Returns a
        DataFrame of loss projections for each subsequent development period
        for each accident year. Populates the triangle's lower-right or
        southeast portion (i.e., "squaring the triangle").

        Returns
        -------
        pd.DataFrame
        """
        if self._trisqrd is None:
            self._trisqrd = self.tri.copy(deep=True)
            ldfs = self.ldfs.values
            rposf = self.tri.index.size
            clvi = self.tri.clvi["row_offset"]
            for i in enumerate(self._trisqrd.columns[1:], start=1):
                ii  , devp  = i[0], i[1]
                ildf, rposi = ldfs[ii - 1], clvi[devp] + 1
                self._trisqrd.iloc[rposi:rposf, ii] = \
                    self._trisqrd.iloc[rposi:rposf, ii - 1] * ildf
            # Multiply right-most column by tail factor.
            max_devp = self._trisqrd.columns[-1]
            self._trisqrd["ultimate"] = self._trisqrd.loc[:,max_devp].values * self.tail
        return(self._trisqrd.astype(np.float_).sort_index())



    def plot(self, actuals_color="#334488", forecasts_color="#FFFFFF",
             axes_style="darkgrid", context="notebook", col_wrap=5,
             **kwargs):
        """
        Visualize projected chain ladder development. First transforms data
        into long format, then plots actual and forecast loss amounts at each
        development period for each origin year using seaborn's ``FacetGrid``.

        Parameters
        ----------
        actuals_color: str
            A hexidecimal color code used to represent actual scatter points
            in FacetGrid. Defaults to "#00264C".

        forecasts_color: str
            A hexidecimal color code used to represent forecast scatter points
            in FacetGrid. Defaults to "#FFFFFF".

        axes_style: str
            Aesthetic style of plots. Defaults to "darkgrid". Other options
            include: {whitegrid, dark, white, ticks}.

        context: str
            Set the plotting context parameters. According to the seaborn
            documentation, This affects things like the size of the labels,
            lines, and other elements of the plot, but not the overall style.
            Defaults to ``"notebook"``. Additional options include
            {paper, talk, poster}.

        col_wrap: int
            The maximum number of origin period axes to have on a single row
            of the resulting FacetGrid. Defaults to 5.

        kwargs: dict
            Additional styling options for scatter points. This can override
            default values for ``plt.plot`` objects. For a demonstration,
            See the Examples section.

        Returns
        -------
        None

        Examples
        --------
        Demonstration of how to pass a dictionary of plot properties in order
        to update the scatter size and marker:

        >>> import trikit
        >>> raa = trikit.load(dataset="raa")
        >>> cl_init = trikit.chladder(data=raa)
        >>> cl_result = cl_init(sel="simple-5", tail=1.005)

        ``cl_result`` represents an instance of ``_ChainLadderResult``, which
        exposes the ``plot`` method. First, we compile the dictionary of
        attributes to override:

        >>> kwds = dict(marker="s", markersize=6)
        >>> cl_result.plot(**kwds)
        """
        df0 = self.trisqrd.reset_index(drop=False).rename({"index":self.tri.origin}, axis=1)
        df0 = pd.melt(df0, id_vars=[self.tri.origin], var_name=self.tri.dev, value_name=self.tri.value)
        df0 = df0[~np.isnan(df0[self.tri.value])].reset_index(drop=True)
        df1 = self.tri.triind.reset_index(drop=False).rename({"index":self.tri.origin}, axis=1)
        df1 = pd.melt(df1, id_vars=[self.tri.origin], var_name=self.tri.dev, value_name=self.tri.value)
        df1[self.tri.value] = df1[self.tri.value].map(lambda v: 1 if v==0 else 0)
        df1 = df1[~np.isnan(df1[self.tri.value])].rename({self.tri.value:"actual_ind"}, axis=1)
        df1 = df1.reset_index(drop=True)
        if self.tail!=1:
            df0[self.tri.dev] = df0[self.tri.dev].map(
                lambda v: (self.tri.devp.max() + 1) if v=="ultimate" else v
                )
        else:
            df0 = df0[df0[self.tri.dev]!="ultimate"]

        # Combine df0 and df1 into a single DataFrame, then perform cleanup
        # actions for cases in which df0 has more records than df1.
        df = pd.merge(df0, df1, on=[self.tri.origin, self.tri.dev], how="left", sort=False)
        df["actual_ind"] = df["actual_ind"].map(lambda v: 0 if np.isnan(v) else v)
        df["actual_ind"] = df["actual_ind"].astype(np.int_)
        df = df.sort_values([self.tri.origin, self.tri.dev]).reset_index(drop=True)
        dfma = df[df["actual_ind"]==1].groupby([self.tri.origin])[self.tri.dev].max().to_frame()
        dfma = dfma.reset_index(drop=False).rename(
            {"index":self.tri.origin, self.tri.dev:"max_actual"}, axis=1)
        df = pd.merge(df, dfma, on=self.tri.origin, how="outer", sort=False)
        df = df.sort_values([self.tri.origin, self.tri.dev]).reset_index(drop=True)
        df["incl_actual"] = df["actual_ind"].map(lambda v: 1 if v==1 else 0)
        df["incl_pred"] = df.apply(
            lambda rec: 1 if (rec.actual_ind==0 or rec.dev==rec.max_actual) else 0,
            axis=1
            )

        # Vertically concatenate dfact_ and dfpred_.
        dfact_ = df[df["incl_actual"]==1][["origin", "dev", "value"]]
        dfact_["description"] = "actual"
        dfpred_ = df[df["incl_pred"]==1][["origin", "dev", "value"]]
        dfpred_["description"] = "forecast"
        data = pd.concat([dfact_, dfpred_]).reset_index(drop=True)

        # Plot chain ladder projections by development period for each
        # origin year. FacetGrid's ``hue`` argument should be set to
        # "description".
        sns.set_context(context)
        with sns.axes_style(axes_style):
            titlestr_ = "Chain Ladder Projections with Actuals by Origin"
            palette_ = dict(actual=actuals_color, forecast=forecasts_color)
            pltkwargs = dict(
                marker="o", markersize=7, alpha=1, markeredgecolor="#000000",
                markeredgewidth=.50, linestyle="--", linewidth=.75,
                fillstyle="full",
                )

            if kwargs:
                pltkwargs.update(kwargs)

            g = sns.FacetGrid(
                data, col="origin", hue="description", palette=palette_,
                col_wrap=col_wrap, margin_titles=False, despine=True, sharex=True,
                sharey=True, hue_order=["forecast", "actual",]
                )

            g.map(plt.plot, "dev", "value", **pltkwargs)
            g.set_axis_labels("", "")
            g.set(xticks=data.dev.unique().tolist())
            g.set_titles("{col_name}", size=9)
            g.set_xticklabels(data.dev.unique().tolist(), size=8)

            # Change ticklabel font size and place legend on each facet.
            for i, _ in enumerate(g.axes):
                ax_ = g.axes[i]
                legend_ = ax_.legend(
                    loc="lower right", fontsize="x-small", frameon=True,
                    fancybox=True, shadow=False, edgecolor="#909090",
                    framealpha=1, markerfirst=True,)
                legend_.get_frame().set_facecolor("#FFFFFF")
                ylabelss_ = [i.get_text() for i in list(ax_.get_yticklabels())]
                ylabelsn_ = [float(i.replace(u"\u2212", "-")) for i in ylabelss_]
                ylabelsn_ = [i for i in ylabelsn_ if i>=0]
                ylabels_ = ["{:,.0f}".format(i) for i in ylabelsn_]
                ax_.set_yticklabels(ylabels_, size=8)

                # Draw border around each facet.
                for _, spine_ in ax_.spines.items():
                    spine_.set_visible(True)
                    spine_.set_color("#000000")
                    spine_.set_linewidth(.50)

        # Adjust facets downward and and left-align figure title.
        plt.subplots_adjust(top=0.9)
        g.fig.suptitle(
            titlestr_, x=0.065, y=.975, fontsize=11, color="#404040", ha="left"
            )


    def __str__(self):
        formats_ = {"ultimate":"{:.0f}".format, "reserve":"{:.0f}".format,
                    "latest":"{:.0f}".format, "cldf":"{:.5f}".format,}
        return(self.summary.to_string(formatters=formats_))


    def __repr__(self):
        formats_ = {"ultimate":"{:.0f}".format, "reserve":"{:.0f}".format,
                    "latest":"{:.0f}".format, "cldf":"{:.5f}".format,}
        return(self.summary.to_string(formatters=formats_))



import sys; sys.path.append("E:\\Repos\\")
import unittest
import trikit
import pandas as pd
import numpy as np
import os
import os.path
import decimal
import logging
import timeit
import matplotlib.pyplot as plt
import seaborn as sns

from trikit.triangle import (
    IncrTriangle,
    CumTriangle,
    _cumtoincr,
    _incrtocum,
    _tritotbl,
    plot_devp
    )

RAA  = trikit.load("raa")
ta83 = trikit.load("ta83")
ctri = trikit.CumTriangle(data=RAA)


tri_init = pd.DataFrame({
    1 :[94,101,82,110,68,119,72,71,71,62],
    2 :[119,131,107,139,99,151,99,101,96,np.NaN],
    3 :[124,135,112,146,105,157,99,106,np.NaN,np.NaN],
    4 :[128,139,116,152,108,158,99,np.NaN,np.NaN,np.NaN],
    5 :[130,141,119,154,111,162,np.NaN,np.NaN,np.NaN,np.NaN],
    6 :[132,143,119,155,114,np.NaN,np.NaN,np.NaN,np.NaN,np.NaN],
    7 :[133,143,120,156,np.NaN,np.NaN,np.NaN,np.NaN,np.NaN,np.NaN],
    8 :[133,144,121,np.NaN,np.NaN,np.NaN,np.NaN,np.NaN,np.NaN,np.NaN],
    9 :[133,145,np.NaN,np.NaN,np.NaN,np.NaN,np.NaN,np.NaN,np.NaN,np.NaN],
    10:[134,np.NaN,np.NaN,np.NaN,np.NaN,np.NaN,np.NaN,np.NaN,np.NaN,np.NaN]
    }, index=range(1991,2001,1)
    )


tt = pd.DataFrame({
    1 :[94,101,82,110,68,119,72,71,71,62],
    2 :[119,131,107,139,99,151,99,101,96,np.NaN],
    3 :[124,135,112,146,105,157,99,106,np.NaN,np.NaN],
    4 :[128,139,116,152,108,158,99,np.NaN,np.NaN,np.NaN],
    5 :[130,141,119,154,111,162,np.NaN,np.NaN,np.NaN,np.NaN],
    6 :[132,143,119,155,114,np.NaN,np.NaN,np.NaN,np.NaN,np.NaN],
    7 :[133,143,120,156,np.NaN,np.NaN,np.NaN,np.NaN,np.NaN,np.NaN],
    8 :[133,144,121,np.NaN,np.NaN,np.NaN,np.NaN,np.NaN,np.NaN,np.NaN],
    9 :[133,145,np.NaN,np.NaN,np.NaN,np.NaN,np.NaN,np.NaN,np.NaN,np.NaN],
    10:[134,np.NaN,np.NaN,np.NaN,np.NaN,np.NaN,np.NaN,np.NaN,np.NaN,np.NaN]
    }, index=range(1991,2001,1)
    )


# Instantiate MackChainLadder instance =======================================>
#btri = trikit.totri(data=tri_init, trifmt="cumulative", tritype="cumulative")
#bcl  = trikit.BootChainLadder(btri)


# ============================================================================>
bcl = trikit.BootChainLadder(RAA, sims=10, procdist="gamma", random_state=30)

nbrcells = bcl.nbr_cells
dof      = bcl.dof
sclparam = bcl.scale_param
trifitcum = bcl.tri_fit_cum
trifitincr = bcl.tri_fit_incr
resid_us    = bcl.resid_us
resid_adj   = bcl.resid_adj
sampd       = bcl.sampling_dist
bs_samps    = bcl.bs_samples
bs_ldfs     = bcl.bs_ldfs
bs_fore     = bcl.bs_forecasts
bspe        = bcl.bs_process_error(all_cols=False)


s = bcl.bootstrap_samples
l = bcl.bootstrap_ldfs
f = bcl.bootstrap_forecasts



import sys; sys.path.append("E:\\Repos\\")
import unittest
import trikit
import pandas as pd
import numpy as np
import os
import os.path
import decimal
import logging
import timeit
import matplotlib.pyplot as plt
import seaborn as sns

from trikit.triangle import (
    IncrTriangle,
    CumTriangle,
    _cumtoincr,
    _incrtocum,
    _tritotbl,
    plot_devp
    )

"""
 ### TODO ###
- Check that cl.ultimates == trisqrd[-1]
- Check that cl.trisqrd - cl.tri == 0
- Check that cl0.ultimates-cl0.reserves-cl0.latest == 0
"""

RAA  = trikit.load("raa")
ta83 = trikit.load("ta83")
ctri = trikit.CumTriangle(data=RAA)



# Instantiate MackChainLadder instance =======================================>
mcl0 = trikit.MackChainLadder(data=ta83, alpha=1, tail=1)
mcl1 = trikit.MackChainLadder(data=RAA, alpha=1, tail=1)

ldfs    = mcl0.ldfs
cldfs   = mcl0.cldfs
invsums = mcl0.inverse_sums
devpvar  = mcl0.devpvar
procerr = mcl0.process_error
originref = mcl0.originref
devpref = mcl0.devpref
rmse    = mcl0.rmsepi

def f(*args):
    return([i for i in args])



"""
_MackChainLadder class definition.
"""
import numpy as np
import pandas as pd
from scipy.stats import norm, lognorm
from . import _BaseChainLadder



class _MackChainLadder(_BaseChainLadder):
    """
    Perform Mack Chain Ladder method. The predicition variance is comprised
    of the estimation variance and the process variance. Estimation
    variance arises from the inability to accurately define the distribution
    from which past events have been generated. Process variance arises from
    the inability to accurately predict which single outcome from the
    distribution will occur at a given time. The predicition error is
    specified as the standard deviation of the forecast.
    """
    def __init__(self, cumtri):
        """
        Parameters
        ----------
        cumtri: triangle._CumTriangle
            A cumulative.CumTriangle instance
        """

        super().__init__(cumtri)

        # properties
        self._parameter_error = None
        self._process_error = None
        self._inverse_sums = None
        self._originref = None
        self._devpref = None
        self._devpvar = None
        self._mseptot = None
        self._rmsepi = None
        self._msepi = None


    def __call__(self, sel="mack-alpha-1", tail=1.0):
        """
        Return a summary of ultimate and reserve estimates resulting from
        the application of the development technique over self.tri. Summary
        DataFrame is comprised of origin year, maturity of origin year, loss
        amount at latest evaluation, cumulative loss development factors,
        projected ultimates and the reserve estimate, by origin year and in
        aggregate.

        Returns
        -------
        pd.DataFrame
            Summary with values by origin year for maturity, latest cumulative
            loss amount, cumulative loss development factor, projected
            ultimate loss and the reserve amount.
        """
        # Bind reference to class method return values.
        ldfs    = self._ldfs(sel=sel, tail=tail)
        cldfs   = self._cldfs(sel=sel, tail=tail)
        ults    = self._ultimates(cldfs=cldfs)
        ibnr    = self._reserves(ultimates=ults)
        sigma_i = np.log(1 + self.msepi / ibnr**2)
        mu_i    = np.log(ibnr - .50 * sigma_i)

        summcols = ["maturity", "latest", "cldf", "ultimate", "reserve"]
        summDF   = pd.DataFrame(columns=summcols, index=self.tri.index)
        summDF["maturity"]  = self.tri.maturity.astype(np.str)
        summDF["latest"]    = self.tri.latest_by_origin
        summDF["cldf"]      = cldfs.values[::-1]
        summDF["ultimate"]  = ults
        summDF["reserve"]   = ibnr
        self._summary['RMSEP']        = self.rmsepi
        self._summary['CV']           = self.rmsepi/self.reserves
        self._summary["NORM_95%_LB"]  = self.reserves - (1.96 * self.rmsepi)
        self._summary["NORM_95%_UB"]  = self.reserves + (1.96 * self.rmsepi)
        self._summary["LNORM_95%_LB"] = np.exp(mu_i - 1.96 * sigma_i)
        self._summary["LNORM_95%_UB"] = np.exp(mu_i + 1.96 * sigma_i)
        self._summary.loc['TOTAL']    = self._summary.sum()
        summDF.loc["total"] = summDF.sum()

        # Set to NaN columns that shouldn't be summed.
        summDF.loc["total", "maturity"] = ""
        summDF.loc["total", "cldf"]     = np.NaN
        summDF = summDF.reset_index().rename({"index":"origin"}, axis="columns")




            # # Populate self._summary with existing properties if available.
            # self._summary['LATEST']       = self.latest_by_origin
            # self._summary['CLDF']         = self.cldfs[::-1]
            # self._summary['EMERGENCE']    = 1/self.cldfs[::-1]
            # self._summary['ULTIMATE']     = self.ultimates
            # self._summary['RESERVE']      = self.reserves
            # self._summary['RMSEP']        = self.rmsepi
            # self._summary['CV']           = self.rmsepi/self.reserves
            # self._summary["NORM_95%_LB"]  = self.reserves - (1.96 * self.rmsepi)
            # self._summary["NORM_95%_UB"]  = self.reserves + (1.96 * self.rmsepi)
            # self._summary["LNORM_95%_LB"] = np.exp(mu_i - 1.96 * sigma_i)
            # self._summary["LNORM_95%_UB"] = np.exp(mu_i + 1.96 * sigma_i)
            # self._summary.loc['TOTAL']    = self._summary.sum()
            #
            # # Set CLDF Total value to `NaN`.
            # self._summary.loc["TOTAL","CLDF","EMERGENCE"] = np.NaN

        return(self._summary)


    @staticmethod
    def get_quantile(*pctl):
        pass




    @property
    def originref(self):
        """
        Intended for internal use only. Contains data by origin year.
        """
        if self._originref is None:
            self._originref = pd.DataFrame({
                'reserve'      :self.reserves,
                'ultimate'     :self.ultimates,
                'process_error':self.process_error,
                'param_error'  :self.parameter_error,
                'msep'         :self.msepi,
                'rmsep'        :self.rmsepi}, index=self.tri.index)
            self._originref = \
                self._originref[
                    ["reserve","ultimate","process_error","param_error","msep","rmsep"]
                    ]
        return(self._originref)



    @property
    def devpref(self):
        """
        Intended for internal use only. Contains data by development period.
        """
        if self._devpref is None:
            self._devpref = pd.DataFrame({
                "ldf"    :self.ldfs,
                "sse"    :self.devpvar.values,
                "ratio"  :(self.devpvar.values / self.ldfs ** 2),
                "dev"    :self.tri.columns[:-1],
                "inv_sum":self.inverse_sums},
                index=self.tri.columns[:-1]
                )

            self._devpref["indx"] = \
                self._devpref["dev"].map(
                    lambda x: self.tri.columns.get_loc(x))

            self._devpref = \
                self._devpref[["dev","indx","ldf","sse","ratio","inv_sum"]]

        return(self._devpref)



    @property
    def inverse_sums(self):
        """
        Convenience aggregation for use in parameter error
        calcuation.
        """
        if self._inverse_sums is None:
            devp_sums = \
                self.tri.sum(axis=0)-self.tri.latest_by_origin[::-1].values
            self._inverse_sums = pd.Series(
                data=devp_sums,index=devp_sums.index,name='inverse_sums')
            self._inverse_sums = (1 / (self._inverse_sums))[:-1]
        return(self._inverse_sums)



    # @property
    # def inverse_sums2(self):
    #     """
    #     Convenience aggregation for use in parameter error
    #     calcuation.
    #     """
    #     if self._inverse_sums is None:
    #         devp_sums = list()
    #         for devp in self.tri.columns[:-1]:
    #             iterep  = self.tri.index.get_loc(self.tri[devp].last_valid_index())
    #             devpos  = self.tri.columns.get_loc(devp)
    #             itersum = self.tri.iloc[:iterep,devpos].sum()
    #             devp_sums.append((devp,(1/itersum)))
    #         indx, vals = zip(*devp_sums)
    #         self._inverse_sums = \
    #             pd.Series(data=vals, index=indx, name='inverse_sums')
    #     return(self._inverse_sums)

    @property
    def devpvar(self) -> np.ndarray:
        """
        devpvar = `development period variance`. Return the variance
        of each n-1 development periods as a Series object.
        """
        if self._devpvar is None:
            n = self.tri.columns.size
            self._devpvar = np.zeros(n - 1, dtype=np.float_)
            for k in range(n - 2):
                iter_ses = 0  # `square of standard error`
                for i in range(n - (k + 1)):
                    c_1, c_2 = self.tri.iloc[i, k], self.tri.iloc[i,k+1]
                    iter_ses+=c_1*((c_2 / c_1) - self.ldfs[k])**2
                iter_ses = iter_ses/(n-k-2)
                self._devpvar[k] = iter_ses

                # Calculate standard error for dev period n-1.
            self._devpvar[-1] = \
                np.min((
                    self._devpvar[-2]**2 / self._devpvar[-3],
                    np.min([self._devpvar[-2],self._devpvar[-3]])
                    ))
            self._devpvar = pd.Series(
                data=self._devpvar, index=self.tri.columns[:self._devpvar.size],
                name="sqrd_std_error")
        return(self._devpvar)




    @property
    def process_error(self):
        """
        Process error (forecast error) calculation. The process error
        component originates from the stochastic movement of the process.
        Returns a pandas Series containing estimates of process variance by
        origin year.
        """
        if self._process_error is None:
            lastcol, pelist = self.tri.columns.size-1, list()
            for rindx in self.tri.rlvi.index:
                iult = self.ultimates[rindx]
                ilvi = self.tri.rlvi.loc[rindx,:].col_offset
                ilvc = self.tri.rlvi.loc[rindx,:].dev
                ipe  = 0
                if ilvi<lastcol:
                    for dev in self.trisqrd.loc[rindx][ilvi:lastcol].index:
                        ipe+=(self.devpref.loc[dev,'RATIO'] / self.trisqrd.loc[rindx,dev])
                    ipe*=(iult**2)
                pelist.append((rindx,ipe))

            # Convert list of tuples into Series object.
            indx, vals = zip(*pelist)
            self._process_error = \
                pd.Series(data=vals, index=indx, name="process_error")
        return(self._process_error)




    @property
    def parameter_error(self):
        """
        Estimation error (parameter error) reflects the uncertainty in
        the estimation of the parameters.
        """
        if self._parameter_error is None:
            lastcol, pelist = self.tri.columns.size-1, list()
            for i in enumerate(self.tri.index):
                ii, rindx = i[0], i[1]
                iult = self.ultimates[rindx]
                ilvi = self.tri.rlvi.loc[rindx,:].col_offset
                ilvc = self.tri.rlvi.loc[rindx,:].dev
                ipe  = 0
                if ilvi<lastcol:
                    for k in range(ilvi,lastcol):
                        ratio  = self.devpref[self.devpref['indx']==k]['ratio'].values[0]
                        invsum = self.devpref[self.devpref['indx']==k]['inv_sum'].values[0]
                        ipe+=(ratio * invsum)
                    ipe*=iult**2
                else:
                    ipe = 0
                pelist.append((rindx, ipe))
            # Convert list of tuples into Series object.
            indx, vals = zip(*pelist)
            self._parameter_error = \
                pd.Series(data=vals, index=indx, name="parameter_error")
        return(self._parameter_error)




    @property
    def covariance_term(self):
        """
        Used to derive the conditional mean squared error of total
        reserve prediction. MSE_(i,j) is non-zero only for cells
        in which i < j (i.e., the
        :return:
        """
        pass


    @property
    def msepi(self):
        """
        Return the mean squared error of predicition by origin year.
        Does not contain estimate for total MSEP.
        MSE_i = process error + parameter error
        """
        if self._msepi is None:
            self._msepi = self.process_error + self.parameter_error
        return(self._msepi)


    @property
    def rmsepi(self):
        """
        Return the root mean squared error of predicition by origin
        year. Does not contain estimate for total MSEP.

            MSE_i = process error + parameter error
        """
        if self._rmsepi is None:
            self._rmsepi = np.sqrt(self.msepi)
        return(self._rmsepi)




    @property
    def summary(self):
        """
        Return a DataFrame containing summary statistics resulting
        from applying the development method to tri, in addition
        to Mack-generated range estimates.
        """
        if self._summary is None:
            self._summary = \
                pd.DataFrame(
                    columns=[
                        "LATEST","CLDF","EMERGENCE","ULTIMATE","RESERVE","RMSEP",
                        "CV","NORM_95%_LB","NORM_95%_UB","LNORM_95%_LB","LNORM_95%_UB"
                        ], index=self.tri.index
                    )

            # Initialize lognormal confidence interval parameters.
            sigma_i = np.log(1 + self.msepi/self.reserves**2)
            mu_i    = np.log(self.reserves - .50 * sigma_i)

            # Populate self._summary with existing properties if available.
            self._summary['LATEST']       = self.latest_by_origin
            self._summary['CLDF']         = self.cldfs[::-1]
            self._summary['EMERGENCE']    = 1/self.cldfs[::-1]
            self._summary['ULTIMATE']     = self.ultimates
            self._summary['RESERVE']      = self.reserves
            self._summary['RMSEP']        = self.rmsepi
            self._summary['CV']           = self.rmsepi/self.reserves
            self._summary["NORM_95%_LB"]  = self.reserves - (1.96 * self.rmsepi)
            self._summary["NORM_95%_UB"]  = self.reserves + (1.96 * self.rmsepi)
            self._summary["LNORM_95%_LB"] = np.exp(mu_i - 1.96 * sigma_i)
            self._summary["LNORM_95%_UB"] = np.exp(mu_i + 1.96 * sigma_i)
            self._summary.loc['TOTAL']    = self._summary.sum()

            # Set CLDF Total value to `NaN`.
            self._summary.loc["TOTAL","CLDF","EMERGENCE"] = np.NaN

        return(self._summary)


    @staticmethod
    def get_quantile(*pctl):
        pass



    def __repr__(self):
        """
        Override default numerical precision used for representing
        ultimate loss projections and age-to-ultimate factors.
        """
        summary_cols = [
            "LATEST","CLDF","EMERGENCE","ULTIMATE","RESERVE","RMSEP",
            "CV","NORM_95%_LB","NORM_95%_UB","LNORM_95%_LB","LNORM_95%_UB"
            ]

        summ_specs = pd.Series([0, 5, 5, 0, 0, 0, 5, 0, 0, 0, 0], index=summary_cols)
        return(self.summary.round(summ_specs).to_string())





import sys; sys.path.insert(0, "E:/Repos/trikit/tests")
from trikit_tests import *


# def _bs_reserves1(pedf):
#
#     grpsdf = pedf.groupby(["sim", "origin"], sort=False)
#
#     dflist = [
#         grpsdf.get_group(i).reset_index(drop=True) for i in grpsdf.groups
#         ]
#
#     for j in range(len(dflist)):
#
#         dflist[j].sort_values(by=["dev"], ascending=True, inplace=True)
#
#         l_act_indx  = dflist[j].index[dflist[j].rectype=="actual"].max()
#         l_fcst_init = dflist[j].index[dflist[j].rectype=="forecast"].max()
#         l_fcst_indx = l_act_indx if np.isnan(l_fcst_init) else l_fcst_init
#         dflist[j].loc[l_fcst_indx, "latest_actual"] = dflist[j].at[l_act_indx, "final_cum"]
#         dflist[j]["reserve"] = dflist[j]["final_cum"] - dflist[j]["latest_actual"]
#     resdf = pd.concat(dflist)
#     return(resdf.sort_values(["sim", "origin", "dev"]).reset_index(drop=True))
#
#
# def _bs_reserves2(pedf):
#     keepcols = ["sim", "origin", "latest", "ultimate", "reserve"]
#     max_devp = pedf["dev"].values.max()
#     pedf = pedf.rename(columns={"final_cum":"ultimate", "l_act_cum":"latest"})
#     pedf["reserve"] = pedf["ultimate"] - pedf["latest"]
#     resvdf = pedf[pedf["dev"].values==max_devp][keepcols].drop_duplicates()
#     #resvdf = resvdf[keepcols].drop_duplicates()
#     resvdf["latest"]  = np.where(np.isnan(resvdf["latest"].values), resvdf["ultimate"].values, resvdf["latest"].values)
#     resvdf["reserve"] = np.nan_to_num(resvdf["reserve"].values, 0)
#     return(resvdf.sort_values(by=["origin", "sim"]).reset_index(drop=True))
#
#
#
#
#
# def _ldfs0(samples):
#
#     df, lvi  = samples, bcl.tri.clvi.reset_index(drop=False)
#
#     lvi.rename(
#         columns={
#             "index":"dev", "origin":"last_origin",
#             "row_offset":"origin_offset"
#             }, inplace=True
#         )
#
#     df = df.merge(lvi, how="left", on=["dev"])
#     dfgrp = df.groupby(["sim"], sort=False)
#     ldflist = list()
#     simlist = [
#         dfgrp.get_group(i).reset_index(drop=True)
#             for i in dfgrp.groups
#         ]
#
#     for indx, simgrp in enumerate(simlist):
#         idf = simgrp[["origin", "dev", "samp_cum", "last_origin"]]
#         idf = idf.reset_index(drop=True)
#         ll  = list()
#         pairs = zip(
#             idf.dev.unique()[::-1][:-1].tolist(),
#             idf.dev.unique()[::-1][1:].tolist()
#             )
#
#         for devp2, devp1 in pairs:
#             devp2sum = idf[idf.dev==devp2].samp_cum.sum()
#             devp2lo  = idf[idf.dev==devp2].last_origin.unique()[0]
#             devp1sum = idf[(idf.dev==devp1) & (idf.origin <= devp2lo)].samp_cum.sum()
#             ll.append((devp1, (devp2sum / devp1sum)))
#
#         idevp, ildfs = zip(*ll)
#         ldflist.append(
#             pd.DataFrame({"dev":idevp, "ldf":ildfs, "sim":indx})
#             )
#
#     _ldfs = pd.concat(ldflist)[["sim", "dev", "ldf"]]
#     return(_ldfs.sort_values(by=["sim", "dev"]).reset_index(drop=True))
#
#
#
# def _ldfs1(samples):
#
#     #samples = _samples
#     lvi = bcl.tri.clvi.reset_index(drop=False)
#
#     lvi.rename(
#         columns={
#             "index":"dev", "origin":"last_origin", "row_offset":"origin_offset"
#             }, inplace=True
#         )
#
#     samples2 = samples.merge(lvi, how="left", on=["dev"])
#     dfgroups = samples2.groupby(["sim"], sort=False)
#     keepcols = ["sim", "origin", "dev", "samp_cum", "last_origin"]
#     ldflist  = []
#
#     for indx, simgrp in dfgroups:
#
#         initdf = simgrp[keepcols].reset_index(drop=True)
#         iterdf = initdf.sort_values(by=["sim", "dev", "origin"], ascending=[True, True, True])
#         iterdf = iterdf[~np.isnan(iterdf.samp_cum)].reset_index(drop=True)
#         iterdf["_aggdev1"]  = iterdf.groupby(["sim", "dev"])["samp_cum"].transform(pd.Series.sum)
#         iterdf["_aggdev2"] = iterdf.apply(lambda rec: 0 if rec.origin==rec.last_origin else rec.samp_cum, axis=1)
#         iterdf["_aggdev2"] = iterdf.groupby(["sim", "dev"])["_aggdev2"].transform(pd.Series.sum)
#         uniqdf = iterdf[["sim", "dev", "_aggdev1", "_aggdev2"]].drop_duplicates().reset_index(drop=True)
#         uniqdf["_aggdev2"] = uniqdf["_aggdev2"].shift(periods=1, axis=0)
#         uniqdf["dev"] = uniqdf["dev"].shift(periods=1, axis=0)
#         uniqdf["ldf"] = uniqdf._aggdev1 / uniqdf._aggdev2
#         ldflist.append(uniqdf)
#
#     _ldfs = pd.concat(ldflist).dropna(how="any").drop(labels=["_aggdev1", "_aggdev2"], axis=1)
#
#     _ldfs["dev"] = _ldfs["dev"].astype(np.integer)
#
#     return(_ldfs.sort_values(["sim", "dev"]).reset_index(drop=True))
#
#
#
# def _ldfs2(samples):
#
#     #samples = _samples
#     keepcols = ["sim", "origin", "dev", "samp_cum", "last_origin"]
#
#     lvi = self.tri.clvi.reset_index(drop=False)
#
#     lvi.rename(
#         columns={
#             "index":"dev", "origin":"last_origin", "row_offset":"origin_offset"
#             }, inplace=True
#         )
#
#     samples2 = samples.merge(lvi, how="left", on=["dev"])
#     initdf = samples2[keepcols].sort_values(by=["sim", "dev", "origin"], ascending=[True, True, True])
#     ldfsdf = initdf[~np.isnan(initdf["samp_cum"])].reset_index(drop=True)
#     ldfsdf["_aggdev1"] = ldfsdf.groupby(["sim", "dev"])["samp_cum"].transform("sum")
#     ldfsdf["_aggdev2"] = np.where(ldfsdf["origin"].values==ldfsdf["last_origin"].values, 0, ldfsdf["samp_cum"].values)
#     ldfsdf["_aggdev2"] = ldfsdf.groupby(["sim", "dev"])["_aggdev2"].transform("sum")
#     uniqdf = ldfsdf[["sim", "dev", "_aggdev1", "_aggdev2"]].drop_duplicates().reset_index(drop=True)
#     uniqdf["_aggdev2"] = uniqdf["_aggdev2"].shift(periods=1, axis=0)
#     uniqdf["dev"] = uniqdf["dev"].shift(periods=1, axis=0)
#     _ldfs = uniqdf[uniqdf["_aggdev2"]!=0].dropna(how="any")
#     _ldfs["ldf"] = _ldfs["_aggdev1"] / _ldfs["_aggdev2"]
#     _ldfs["dev"] = _ldfs["dev"].astype(np.integer)
#     return(_ldfs[["sim", "dev", "ldf"]].reset_index(drop=True))
#
#
# # _bs_forecasts ==============================================================]
#
#
# def _forecasts0(samples, ldfs):
#
#     fulldf = samples.merge(ldfs, how="left", on=["sim", "dev"])
#     grpsdf = fulldf.groupby(["sim", "origin"], sort=False)
#     dflist = [
#         grpsdf.get_group(i).reset_index(drop=True)
#             for i in grpsdf.groups
#         ]
#
#     for j in range(len(dflist)):
#         dflist[j].sort_values(by=["dev"], ascending=True, inplace=True)
#         if np.any(dflist[j].rectype=="forecast"):
#             l_act_indx  = dflist[j].index[dflist[j].rectype=="actual"].max()
#             f_fcst_indx = dflist[j].index[dflist[j].rectype=="forecast"].min()
#             l_act_cum   = dflist[j].loc[l_act_indx, "samp_cum"]
#             dflist[j].loc[f_fcst_indx:, "samp_cum"] = \
#                 dflist[j].loc[l_act_indx:, "ldf"].cumprod().values[:-1] * l_act_cum
#             dflist[j]["samp_incr"][1:] = dflist[j]["samp_cum"].diff(periods=1)[1:]
#     sqrddf = pd.concat(dflist)
#     sqrddf["var"] = \
#         sqrddf.apply(
#             lambda rec: np.abs(rec.samp_incr * 632) \
#                 if rec.rectype=="forecast" else np.NaN, axis=1)
#     sqrddf["sign"] = \
#         sqrddf.apply(
#             lambda rec: 1 if rec.samp_incr > 0 else -1, axis=1)
#     return(sqrddf.sort_values(by=["sim", "origin", "dev"]).reset_index(drop=True))
#
#
#
#
# def _forecasts1a(samples, ldfs):
#
#     fulldf = samples.merge(ldfs, how="left", on=["sim", "dev"]).sort_values(by=["sim", "origin", "dev"])
#     fulldf["_fcst_test"] = fulldf.groupby(["sim", "origin"])["rectype"].transform(lambda v: np.any(v.values=="forecast"))
#     actsdf = fulldf[fulldf["_fcst_test"]==False]; df = fulldf[fulldf["_fcst_test"]==True]
#     df["_l_act_indx"]    = df.groupby(["sim", "origin"])["samp_cum"].transform(lambda v: v.last_valid_index())
#     df["_fcst_ind"]      = (df["_l_act_indx"].values <= df.index)*1
#     df["_l_act_cum"]     = df.groupby(["sim", "origin"])["samp_cum"].transform(lambda v: v[v.last_valid_index()])
#     df["_revldf"]        = np.nan_to_num(np.where(df["_fcst_ind"].values==1, df["ldf"].values, 1), 0)
#     df["_cumldf"]        = df.groupby(["sim", "origin"])["_revldf"].transform(lambda v: v.cumprod()).shift(periods=1, axis=0)
#     df["_samp_cum2"]     = np.nan_to_num((df["_l_act_cum"].values * df["_fcst_ind"].values * df["_cumldf"].values), 0)
#     df["samp_cum_final"] = np.nan_to_num(df["samp_cum"].values * (1 - df["_fcst_ind"].values), 0) + \
#                            np.nan_to_num(df["_samp_cum2"].values * df["_fcst_ind"].values, 0)
#
#     # Drop and rename columns.
#     df.drop(labels=["samp_cum", "samp_incr"], axis=1, inplace=True)
#     df.drop(labels=[i for i in df.columns if i.startswith("_")], axis=1, inplace=True)
#     df.rename(columns={"samp_cum_final":"samp_cum"}, inplace=True)
#     df["samp_incr"] = df["samp_cum"]
#     df["samp_incr"][1:] = df["samp_cum"].diff(periods=1)[1:]
#     actsdf.drop(labels="_fcst_test", axis=1, inplace=True)
#     sqrddf = pd.concat([df, actsdf], sort=True)
#     sqrddf["var"]  = np.where(sqrddf["rectype"].values=="forecast", np.abs(sqrddf["samp_incr"].values * 632.3368030912758), np.NaN)
#     sqrddf["sign"] = np.where(sqrddf["samp_incr"].values > 0, 1, -1)
#     return(sqrddf.sort_values(by=["sim", "origin", "dev"]).reset_index(drop=True))
#
#
#
#
#
# def _forecasts1b(samples):
#     fulldf = _samples.merge(_ldfs, how="left", on=["sim", "dev"]).sort_values(by=["sim", "origin", "dev"])
#     #fulldf = samples
#     fulldf["_fcst_test"] = fulldf.groupby(["sim", "origin"])["rectype"].transform(lambda v: np.any(v.values=="forecast"))
#     actsdf = fulldf[fulldf["_fcst_test"]==False]
#     df = fulldf[fulldf["_fcst_test"]==True]
#
#     df["_l_act_indx"]    = df.groupby(["sim", "origin"])["samp_cum"].transform(lambda v: v.last_valid_index())
#     df["_fcst_ind"]      = (df["_l_act_indx"].values <= df.index)*1
#     df["_l_act_cum"]     = df.groupby(["sim", "origin"])["samp_cum"].transform(lambda v: v[v.last_valid_index()])
#     df["_revldf"]        = np.nan_to_num(np.where(df["_fcst_ind"].values==1, df["ldf"].values, 1), 0)
#     df["_cumldf"]        = df.groupby(["sim", "origin"])["_revldf"].transform(lambda v: v.cumprod()).shift(periods=1, axis=0)
#     df["_samp_cum2"]     = np.nan_to_num((df["_l_act_cum"].values * df["_fcst_ind"].values * df["_cumldf"].values), 0)
#     df["samp_cum_final"] = np.nan_to_num(df["samp_cum"].values * (1 - df["_fcst_ind"].values), 0) + \
#                            np.nan_to_num(df["_samp_cum2"].values * df["_fcst_ind"].values, 0)
#
#     # Drop and rename columns.
#     df.drop(labels=["samp_cum", "samp_incr"], axis=1, inplace=True)
#     df.drop(labels=[i for i in df.columns if i.startswith("_")], axis=1, inplace=True)
#     df.rename(columns={"samp_cum_final":"samp_cum"}, inplace=True)
#     df["samp_incr"] = df["samp_cum"]
#     df["samp_incr"][1:] = df["samp_cum"].diff(periods=1)[1:]
#     actsdf.drop(labels="_fcst_test", axis=1, inplace=True)
#     sqrddf = pd.concat([df, actsdf], sort=True)
#     sqrddf["var"]  = np.where(sqrddf["rectype"].values=="forecast", np.abs(sqrddf["samp_incr"].values * 632.3368030912758), np.NaN)
#     sqrddf["sign"] = np.where(sqrddf["samp_incr"].values > 0, 1, -1)
#     return(sqrddf.sort_values(by=["sim", "origin", "dev"]).reset_index(drop=True))
#
#
# # Alternative approach:
# # Filter out all actual cells prior to performing ultimate calculation
#
#
#
# def _forecasts1c(df):
#
#     #df = combined
#
#     min_origin_year = df["origin"].values.min()
#     df["_l_init_indx"] = np.where(df["dev"].values>=df["_l_act_dev"].values, df.index.values, -1)
#     actsdf = df[(df["origin"].values==min_origin_year) | (df["_l_init_indx"].values==-1)]
#     fcstdf = df[~df.index.isin(actsdf.index)].sort_values(by=["sim", "origin", "dev"])
#     fcstdf["_l_act_indx"] = fcstdf.groupby(["sim", "origin"])["_l_init_indx"].transform("min")
#     fcstdf["_l_act_cum"]  = fcstdf.lookup(fcstdf["_l_act_indx"].values, ["samp_cum"] * fcstdf.shape[0])
#     fcstdf["_cum_ldf"]    = fcstdf.groupby(["sim", "origin"])["ldf"].transform("cumprod").shift(periods=1, axis=0)
#     fcstdf["_samp_cum2"]  = np.nan_to_num((fcstdf["_l_act_cum"].values * fcstdf["_cum_ldf"].values), 0)
#     fcstdf["cum_final"]   = np.nan_to_num(fcstdf["samp_cum"].values, 0) + fcstdf["_samp_cum2"].values
#
#     # Drop and rename columns.
#     fcstdf.drop(labels=["samp_cum", "samp_incr"], axis=1, inplace=True)
#     fcstdf.rename(columns={"cum_final":"samp_cum"}, inplace=True)
#
#     sqrddf = pd.concat([fcstdf, actsdf], sort=True).sort_values(by=["sim", "origin", "dev"])
#     sqrddf["_incr_dev1"] = np.nan_to_num(np.where(sqrddf["dev"].values==1, sqrddf["samp_cum"].values, np.NaN), 0)
#     sqrddf["_incr_dev2"] = np.nan_to_num(sqrddf.groupby(["sim", "origin"])["samp_cum"].diff(periods=1), 0)
#     sqrddf["samp_incr"]  = sqrddf["_incr_dev1"].values + sqrddf["_incr_dev2"].values
#     sqrddf["var"]        = np.abs(sqrddf["samp_incr"].values * 632.3368030912758)
#     sqrddf["sign"]       = np.where(sqrddf["samp_incr"].values > 0, 1, -1)
#     sqrddf.drop(labels=[i for i in sqrddf.columns if i.startswith("_")], axis=1, inplace=True)
#
#     colorder = [
#         "sim", "origin", "dev","incr", "incr_sqrt", "rectype",
#         "resid", "samp_incr", "samp_cum", "ldf", "var", "sign"
#         ]
#
#     return(sqrddf[colorder].sort_values(by=["sim", "origin", "dev"]).reset_index(drop=True))
#
#
#
# # _bs_process_error ==========================================================]
#
# def _bspe1(fcstdf, procdist="gamma", random_state=None):
#
#     #forceasts = fcst2#forecasts
#     procdist = "gamma"
#     random_state = None
#
#     # Initialize pseudo random number generator.
#     if random_state is not None:
#         if isinstance(random_state, int):
#             prng = RandomState(random_state)
#         elif isinstance(random_state, RandomState):
#             prng = random_state
#     else:
#         prng = RandomState()
#
#     # Parameterize distribution to incorporate process variance.
#     if procdist.strip().lower()=="gamma":
#         fcstdf["param2"] = 632.3368030912758
#         fcstdf["param1"] = fcstdf.apply(lambda rec: np.abs(rec.samp_incr / rec.param2) if rec.rectype=="forecast" else np.NaN, axis=1)
#
#         def fdist(param1, param2):
#             """gamma.rvs(a=param1, scale=param2, size=1, random_state=None)"""
#             return(prng.gamma(param1, param2))
#
#     elif procdist.strip().lower()=="odp":
#          # For odp: param1=n (`n`), param2=prob (`p`).
#         fcstdf["param2"] = ((632.3368030912758 - 1) / 632.3368030912758)
#         fcstdf["param1"] = fcstdf.apply(lambda rec: ((rec.samp_incr * (1 - rec.param2)) / rec.param2) if rec.rectype=="forecast" else np.NaN, axis=1)
#
#         def fdist(param1, param2):
#             """nbinom.rvs(n=param1, p=param2, size=1, random_state=None)"""
#             return(prng.nbinom(param1, param2))
#
#     grpsdf = fcstdf.groupby(["sim","origin"], sort=False)
#     dflist = [grpsdf.get_group(i).reset_index(drop=True) for i in grpsdf.groups]
#
#     for j in range(len(dflist)):
#         dflist[j]["final_incr"] = dflist[j].apply(lambda rec: fdist(rec.param1, rec.param2) * rec.sign if rec.rectype=="forecast" else rec.samp_incr, axis=1)
#
#     # Concatenate elements of dflist; compute final cumulative sum.
#     ultdf = pd.concat(dflist).sort_values(by=["sim", "origin", "dev"]).reset_index(drop=True)
#     ultdf["final_cum"] = ultdf.groupby(["sim", "origin"])["final_incr"].cumsum()
#     return(ultdf.sort_values(by=["sim", "origin", "dev"]).reset_index(drop=True))
#
#
#
#
#
# def _bspe2(fcstdf, procdist="gamma", random_state=None):
#
#     # Initialize pseudo random number generator.
#     if random_state is not None:
#         if isinstance(random_state, int):
#             prng = RandomState(random_state)
#         elif isinstance(random_state, RandomState):
#             prng = random_state
#     else:
#         prng = RandomState()
#
#     # Parameterize distribution to incorporate process variance.
#     if procdist.strip().lower()=="gamma":
#         fcstdf["param2"] = 632.3368030912758
#         fcstdf["param1"]  = np.abs(fcstdf["samp_incr"].values / fcstdf["param2"].values)
#
#         def fdist(param1, param2):
#             """gamma.rvs(a=param1, scale=param2, size=1, random_state=None)"""
#             return(prng.gamma(param1, param2))
#
#     elif procdist.strip().lower()=="odp":
#          # For odp: param1=n (`n`), param2=prob (`p`).
#         fcstdf["param2"] = ((632.3368030912758 - 1) / 632.3368030912758)
#         fcstdf["param1"] = (fcstdf["samp_incr"].values * (1 - fcstdf["param2"].values)) / fcstdf["param2"].values
#
#         def fdist(param1, param2):
#             """nbinom.rvs(n=param1, p=param2, size=1, random_state=None)"""
#             return(prng.nbinom(param1, param2))
#
#     fcstdf["final_incr"] = np.where(
#         fcstdf["rectype"].values=="forecast",
#         fdist(fcstdf["param1"].values, fcstdf["param2"].values) * fcstdf["sign"].values,
#         fcstdf["samp_incr"].values
#         )
#
#     fcstdf["final_cum"]  = fcstdf.groupby(["sim", "origin"])["final_incr"].cumsum()
#
#     return(fcstdf.sort_values(by=["sim", "origin", "dev"]).reset_index(drop=True))
#
#


###############################################################################
SIMS          = 100
PROCDIST      = "gamma"
PARAMETRIC    = False
NEG_HANDLER   = 1
PERCENTILES   = [.75, .95]
INTERPOLATION = "linear"
PRNG          = RandomState(516)

bcl_init = trikit.chladder(data=DATA, range_method="bootstrap", neg_handler=1)


bcl = bcl_init(
    sims=SIMS, procdist=PROCDIST, parametric=PARAMETRIC,
    percentiles=PERCENTILES, interpolation=INTERPOLATION, random_state=PRNG
    )

bcl.plotdist(level="agg", tc="#FE0000")




# sns.set(rc={'axes.facecolor':"#1f77b4"})
# g = sns.FacetGrid(dat, col="origin", col_wrap=4, margin_titles=False)
# g.map(plt.hist, "reserve", **plt_params)
# #g.fig.subplots_adjust(wspace=.05, hspace=.05)
# g.set_titles("{col_name}", color="red")
# g.fig.suptitle("Reserve Distribution by Origin Year", color="red", weight="bold")
# plt.subplots_adjust(top=0.92)
#
# plt.show()





tips = sns.load_dataset("tips")
g = sns.FacetGrid(tips, row="sex", col="time", margin_titles=True)
bins = np.linspace(0, 60, 13)
g.map(plt.hist, "total_bill", color="steelblue", bins=bins)


# bcl        = trikit.chladder(data=DATA, range_method="bootstrap", neg_handler=1)
# samples    = bcl._bs_samples(sims=SIMS, parametric=PARAMETRIC, random_state=PRNG)
# ldfs       = bcl._bs_ldfs(samples_df=samples)
# clvi       = bcl.tri.clvi.reset_index(drop=False)
# rlvi       = bcl.tri.rlvi.reset_index(drop=False).rename({"index":"origin", "dev":"l_act_dev"},axis=1).drop(labels="col_offset", axis=1)
# combined   = samples.merge(ldfs, how="left", on=["sim", "dev"]).sort_values(by=["sim", "origin", "dev"])
# combined   = combined.merge(rlvi, how="left", on=["origin"]).reset_index(drop=True).sort_values(by=["sim", "origin", "dev"])
# forecasts  = bcl._bs_forecasts(combined_df=combined)
# pedf       = bcl._bs_process_error(forecasts_df=forecasts, procdist="gamma", random_state=PRNG)
# reserves   = bcl._bs_reserves(process_error_df=pedf)






%load_ext line_profiler
%load_ext memory_profiler

# sims = 1000 ==================]
# _bs_samples       : 162-168
# _bs_ldfs          : 136-141
# _bs_forecasts     : 2472-2528
# _bs_process_error : 75-80

# %lprun -s -T E:/Dev/Profiling/l_forecasts1b.txt -f _forecasts1b _forecasts1b(samples=inputdf)
# %lprun -s -T E:/Dev/Profiling/l_forecasts1a.txt -f _forecasts1a _forecasts1a(samples=_samples, ldfs=_ldfs)
#
# %lprun -s -u 1 -f _forecasts1c _forecasts1c(df=combined)
# %lprun -s -u 1 -f _forecasts1c _forecasts1c(df=combined)

tri = bcl_init.tri.copy(deep=True)

ltri = [tri[[i]].copy() for i in tri.columns]
for i in range(len(ltri)):
    iterdf = ltri[i]
    iterdf.columns.name = None
    devp = iterdf.columns[0]
    iterdf.reset_index(drop=False, inplace=True)
    iterdf.rename(columns={"index":"origin", devp:"value"}, inplace=True)
    iterdf["dev"] = devp
    ltri[i] = iterdf
df = pd.concat(ltri, ignore_index=True)
df = df[~np.isnan(df["value"])]
df = df.sort_values(by=["origin", "dev"]).reset_index(drop=True)
dd1 = df[["origin", "dev", "value"]]

del df, ltri, iterdf

tri = bcl_init.tri.copy(deep=True)
tri = tri.reset_index(drop=False).rename({"index":"origin"}, axis=1)
df = pd.melt(tri, id_vars=["origin"], var_name="dev", value_name="value")
df = df[~np.isnan(df["value"])]
dd2 = df.sort_values(by=["origin", "dev"]).reset_index(drop=True)





"""
trikit sample datasets. Includes Casualty Actuarial Society's Loss Reserving
Database.
The list of available datasets will continue to grow with each release.
Please check back periodically.
"""
import os.path

datasets_dir = os.path.dirname(__file__)

dataref = {
    "nonstd": datasets_dir + os.path.sep + "nonstd.csv",
    "raa"   : datasets_dir + os.path.sep + "RAA.csv",
    "ta83"  : datasets_dir + os.path.sep + "TaylorAshe83.csv",
    "lrdb"  : datasets_dir + os.path.sep + "lrdb.csv",
    }




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="trikit",
    version="0.2.5",
    author="James D. Triveri",
    author_email="james.triveri@gmail.com",
    description="Actuarial Reserving Methods in Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/trikit/trikit",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        ],
    install_requires=[
        "numpy>=1.*", "scipy>=0.19", "pandas>=0.20", "matplotlib>=2.*",
        "seaborn>=0.7",
        ],
    include_package_data=True,
    )



import sys; sys.path.append("G:\\Repos\\")
import unittest
import pandas as pd
import numpy as np
import os
import os.path
import decimal
import logging
import timeit
import matplotlib.pyplot as plt
import seaborn as sns
from scipy import stats
import profile
import cProfile
import io
import pstats
import trikit
from numpy.random import RandomState


TRIANGLE_TEST = False
DATASETS_TEST = False
BASE_CL_TEST  = False
BOOT_CL_TEST  = False
MACK_CL_TEST  = False
MCMC_CL_TEST  = False


triDF0 = pd.DataFrame({
    1 :[94,101,82,110,68,119,72,71,71,62],
    2 :[119,131,107,139,99,151,99,101,96,np.NaN],
    3 :[124,135,112,146,105,157,99,106,np.NaN,np.NaN],
    4 :[128,139,116,152,108,158,99,np.NaN,np.NaN,np.NaN],
    5 :[130,141,119,154,111,162,np.NaN,np.NaN,np.NaN,np.NaN],
    6 :[132,143,119,155,114,np.NaN,np.NaN,np.NaN,np.NaN,np.NaN],
    7 :[133,143,120,156,np.NaN,np.NaN,np.NaN,np.NaN,np.NaN,np.NaN],
    8 :[133,144,121,np.NaN,np.NaN,np.NaN,np.NaN,np.NaN,np.NaN,np.NaN],
    9 :[133,145,np.NaN,np.NaN,np.NaN,np.NaN,np.NaN,np.NaN,np.NaN,np.NaN],
    10:[134,np.NaN,np.NaN,np.NaN,np.NaN,np.NaN,np.NaN,np.NaN,np.NaN,np.NaN]
    }, index=range(1991,2001,1)
    )

df = pd.DataFrame({
    "group" : ["a", "a", "b", "b", "c"],
    "gender": ["M", "F", "F", "M", "F"],
    "score" : np.random.rand(5),
    "salary": [90000, 70000, 74000, 89000, 77000]
    })
df1 = pd.get_dummies(df, columns=["group", "gender"], prefix_sep="_", drop_first=True)
triDF1 = pd.DataFrame({
    "1":[1065,1042,1334,1177,1732],
    "2":[1493,1332,1887,1465,np.NaN],
    "3":[1201,1129,1465,np.NaN,np.NaN],
    "4":[966,1074,np.NaN,np.NaN,np.NaN],
    "5":[0,np.NaN,np.NaN,np.NaN,np.NaN]
    }, index=range(2014,2019)
    )


# Test load function =========================================================]
if DATASETS_TEST:

    lrdb = pd.read_csv("E:/Repos/trikit/datasets/lrdb.csv", sep=",")
    raa  = trikit.load(dataset="raa")
    ta83 = trikit.load(dataset="ta83")
    datasets = trikit.get_datasets()


    # Test calling _get_lrdb_groups with list, dict and pd.DataFrame.
    grps0 = trikit.get_lrdb_groups(returnas=pd.DataFrame)
    grps1 = trikit.get_lrdb_groups(returnas=dict)
    grps2 = trikit.get_lrdb_groups(returnas=list)
    lobs = trikit.get_lrdb_lobs()
    specs = trikit.get_lrdb_specs()


    dat1 = trikit.load(dataset="lrdb", grcode=2143, action=None)
    dat2 = trikit.load(dataset="lrdb", grcode=2143, action="random")
    dat3 = trikit.load(dataset="lrdb", grcode=2143, action="aggregate")

    # Call each lob with varying `action` values.
    dat4a = trikit.load(dataset="lrdb", lob="WC", action=None)
    dat4b = trikit.load(dataset="lrdb", lob="WC", action="agg")
    dat4c = trikit.load(dataset="lrdb", lob="WC", action="rand")

    # Return single group with only `action` specified.
    dat5a = trikit.load(dataset="lrdb", action="random")
    dat5b = trikit.load(dataset="lrdb", lower_right_ind=True, action="random")

    # Varying loss_type parameter.
    dat6a = trikit.load(dataset="lrdb", grcode=2003, loss_type="incurred", action="random")
    dat6b = trikit.load(dataset="lrdb", grcode=3240, loss_type="paid", action="random")
    dat6c = trikit.load(dataset="lrdb", grcode=715, loss_type="bulk", action="random")

    # Filter to full lob while varying `action`.
    # lob1a = trikit.load(dataset="lrdb", lob="WC"       , action="agg")
    # lob2a = trikit.load(dataset="lrdb", lob="PROD_LIAB", action="agg")
    # lob3a = trikit.load(dataset="lrdb", lob="PP_AUTO"  , action="agg")
    # lob4a = trikit.load(dataset="lrdb", lob="OTHR_LIAB", action="agg")
    # lob5a = trikit.load(dataset="lrdb", lob="MED_MAL"  , action="agg")
    # lob6a = trikit.load(dataset="lrdb", lob="COM_AUTO" , action="agg")
    # lob1b = trikit.load(dataset="lrdb", lob="WC"       , action="random")
    # lob2b = trikit.load(dataset="lrdb", lob="PROD_LIAB", action="random")
    # lob3b = trikit.load(dataset="lrdb", lob="PP_AUTO"  , action="random")
    # lob4b = trikit.load(dataset="lrdb", lob="OTHR_LIAB", action="random")
    # lob5b = trikit.load(dataset="lrdb", lob="MED_MAL"  , action="random")
    # lob6b = trikit.load(dataset="lrdb", lob="COM_AUTO" , action="random")


    ld0 = trikit.load(dataset="lrdb", loss_type="paid", action="aggregate")
    ld1 = trikit.load(dataset="lrdb", loss_type="paid", action=None)
    ld2 = trikit.load(dataset="lrdb", loss_type="paid", action="random")
    ld4 = trikit.load(dataset="lrdb", loss_type="paid", action="rand", allcols=False)


# Verify other parameters are ignored when dataset!="lrdb"




# totri testing ==============================================================]
# totri(data, type_="cumulative", origin=None, dev=None, value=None,
#       trifmt=None, datafmt="incremental", **kwargs)
# ============================================================================]
raa  = trikit.load(dataset="raa")
DATA = raa


# ti = trikit.triangle._IncrTriangle(data=DATA)
# tc = trikit.triangle._CumTriangle(data=DATA)
# tc.plot(facets=True)



if TRIANGLE_TEST:

    ##### Case 0 ##############################################################
    # data   :raa
    # type_  :incr
    # origin :None
    # dev    :None
    # value  :None
    # trifmt :None
    # datafmt:incr
    tri0 = trikit.totri(
        data=DATA, type_="incremental", origin=None, dev=None, value=None,
        datafmt="incremental"
        )

    rlvi    = tri0.rlvi
    clvi    = tri0.clvi
    latest  = tri0.latest
    lbo     = tri0.latest_by_origin
    lbd     = tri0.latest_by_devp
    devp    = tri0.devp
    origins = tri0.origins


    ##### Case 1 ##############################################################
    # data   :raa
    # type_  :"cum"
    # origin :None
    # dev    :None
    # value  :None
    # trifmt :None
    # datafmt:"incr"
    tri1 = trikit.totri(
        data=DATA, type_="cumulative", origin=None, dev=None, value=None,
        datafmt="incremental"
        )

    rlvi    = tri1.rlvi
    clvi    = tri1.clvi
    latest  = tri1.latest
    lbo     = tri1.latest_by_origin
    lbd     = tri1.latest_by_devp
    devp    = tri1.devp
    origins = tri1.origins
    linkr   = tri1.a2a
    linkra1  = tri1.a2a_avgs
    linkra2  = tri1.a2a_avgs
    matur   = tri1.maturity

    ##### Case 2 ##############################################################
    # data   :raacum
    # type_  :cum
    # origin :None
    # dev    :None
    # value  :None
    # trifmt :None
    # datafmt:cum
    # tri2 = tri1.cumsum(axis=1)

    raacum = trikit._tritotbl(tri1)
    tri2 = trikit.totri(data=raacum, type_="cumulative", datafmt="cumulative")

    rlvi    = tri2.rlvi
    clvi    = tri2.clvi
    latest  = tri2.latest
    lbo     = tri2.latest_by_origin
    lbd     = tri2.latest_by_devp
    devp    = tri2.devp
    origins = tri2.origins
    linkr   = tri2.a2a
    linkra  = tri2.a2a_avgs
    matur   = tri1.maturity


    ##### Case 3 ##############################################################
    # data   :raacum
    # type_  :cum
    # origin :None
    # dev    :None
    # value  :None
    # trifmt :None
    # datafmt:incr
    raacum = tri1.as_tbl()
    tri3 = trikit.totri(data=raacum, type_="incremental", datafmt="cumulative")

    rlvi    = tri3.rlvi
    clvi    = tri3.clvi
    latest  = tri3.latest
    lbo     = tri3.latest_by_origin
    lbd     = tri3.latest_by_devp
    devp    = tri3.devp
    origins = tri3.origins
    matur   = tri1.maturity


    ##### Case 4 ##############################################################
    # data   :raa
    # type_  :cum
    # origin :None
    # dev    :None
    # value  :None
    # trifmt :None
    # datafmt:""
    cumtri = pd.DataFrame(tri1)
    tri4 = trikit.totri(data=cumtri, type_="cumulative", trifmt="cumulative")

    rlvi    = tri4.rlvi
    clvi    = tri4.clvi
    latest  = tri4.latest
    lbo     = tri4.latest_by_origin
    lbd     = tri4.latest_by_devp
    devp    = tri4.devp
    origins = tri4.origins
    linkr   = tri4.a2a
    linkra  = tri4.a2a_avgs
    matur   = tri1.maturity


    ##### Case 5 ##############################################################
    # data   :raa
    # type_  :cum
    # origin :None
    # dev    :None
    # value  :None
    # trifmt :None
    # datafmt:""
    cumtri = pd.DataFrame(tri1)
    tri5 = trikit.totri(
        data=cumtri, type_="incremental", trifmt="cumulative"
        )

    rlvi    = tri5.rlvi
    clvi    = tri5.clvi
    latest  = tri5.latest
    lbo     = tri5.latest_by_origin
    lbd     = tri5.latest_by_devp
    devp    = tri5.devp
    origins = tri5.origins
    matur   = tri1.maturity


    ##### Case 6 ##############################################################
    # data   :raa
    # type_  :incr
    # origin :None
    # dev    :None
    # value  :None
    # trifmt :None
    # datafmt:""
    incrtri = pd.DataFrame(tri0)
    tri6 = trikit.totri(
        data=incrtri, type_="incremental", trifmt="incremental"
        )

    rlvi    = tri6.rlvi
    clvi    = tri6.clvi
    latest  = tri6.latest
    lbo     = tri6.latest_by_origin
    lbd     = tri6.latest_by_devp
    devp    = tri6.devp
    origins = tri6.origins
    matur   = tri1.maturity


    ##### Case 7 ##############################################################
    # data   :raa
    # type_  :incr
    # origin :None
    # dev    :None
    # value  :None
    # trifmt :None
    # datafmt:""
    incrtri = pd.DataFrame(tri0)
    tri7 = trikit.totri(
        data=incrtri, type_="cumulative", trifmt="incremental"
        )

    rlvi    = tri7.rlvi
    clvi    = tri7.clvi
    latest  = tri7.latest
    lbo     = tri7.latest_by_origin
    lbd     = tri7.latest_by_devp
    devp    = tri7.devp
    origins = tri7.origins
    linkr   = tri7.a2a
    linkra  = tri7.a2a_avgs
    matur   = tri1.maturity



cl_init = trikit.chladder(data=DATA)
ldfs = cl_init._ldfs(sel="all-weighted", tail=1.)
cldfs = cl_init._cldfs(ldfs=ldfs)
ults = cl_init._ultimates(cldfs=cldfs)
res = cl_init._reserves(ultimates=ults)
# r = cl_init(sel="all-weighted", tail=1.0)
#
# cldfs0a = r.cldfs
# ults0a  = r.ultimates
# ibnr0a  = r.reserves
# tsqr0a  = r.trisqrd

# BootstrapChainLadder =======================================================]
bb = trikit.chladder(data=DATA, range_method="bootstrap")

r0 = bb.run(sims=175, neg_handler=1, parametric=False, random_state=516)
r0.plotdist(level="year")


# from trikit.chainladder import bootstrap
# bb = bootstrap._BootstrapChainLadder(cumtri=tri1)

#
# dof = bb.tri.dof
# tfc = bb._tri_fit_cum(sel="all-weighted")
# tfi = bb._tri_fit_incr(fitted_tri_cum=tfc)
# r_us = bb._resid_us(tfi)
# r_adj = bb._resid_adj(resid_us=r_us)
# sclp = bb._scale_param(r_us)
# sampling_dist = bb._sampling_dist(r_adj)
#
# dfsamples = bb._bs_samples(
#     sampling_dist=sampling_dist, fitted_tri_incr=tfi, sims=10, neg_handler=1,
#     parametric=False, random_state=516
#     )
#
# dfldfs = bb._bs_ldfs(dfsamples=dfsamples)
#
# rlvi_ = bb.tri.rlvi.reset_index(drop=False).rename({"index":"origin", "dev":"l_act_dev"}, axis=1)
# dflvi = rlvi_.drop("col_offset", axis=1)
# dfcombined = dfsamples.merge(dfldfs, on=["sim", "dev"], how="left")
# dfcombined = dfcombined.merge(dflvi, how="left", on=["origin"])
# dfcombined = dfcombined.reset_index(drop=True).sort_values(by=["sim", "origin", "dev"])
# dfforecasts = bb._bs_forecasts(dfcombined=dfcombined, scale_param=sclp)
# dfprocerr = bb._bs_process_error(dfforecasts=dfforecasts, scale_param=sclp, procdist="gamma", random_state=516)
# dfreserves = bb._bs_reserves(dfprocerror=dfprocerr)




# Testing _BaseChainLadder  ==================================================]
#
# def chladder(data, origin=None, dev=None, value=None, trifmt=None,
#              datafmt="incremental", tail=1.0, sel="all-weighted",
#              range_method=None, **kwargs)
#
# ============================================================================]

if BASE_CL_TEST:
    # Passing tabular dataset with fields "origin", "dev", "value"
    cl0     = trikit.chladder(data=DATA)
    ldfs0a  = cl0.ldfs
    cldfs0a = cl0.cldfs
    ults0a  = cl0.ultimates
    ibnr0a  = cl0.reserves
    tsqr0a  = cl0.trisqrd

    # Change ldf at `1-2` index; verify change flows through to other properties,
    # starting with self.cldfs > self.ultimates > self.reserves.
    cl0.ldfs = ("1-2", 2.15)

    # Attempt to update non-existent index (results in no-op).
    cl0.ldfs = ("15-16", 1.0034)
    ldfs0b   = cl0.ldfs
    cldfs0b  = cl0.cldfs
    ults0b   = cl0.ultimates
    ibnr0b   = cl0.reserves
    tsqr0b   = cl0.trisqrd

    assert not ldfs0a.equals(ldfs0b)
    assert not cldfs0a.equals(cldfs0b)
    assert not ults0a.equals(ults0b)
    assert not ibnr0a.equals(ibnr0b)
    assert not tsqr0a.equals(tsqr0b)


    # Passing data formatted as cumulative triangle but not typed as such.
    dfc = pd.DataFrame(tri1)
    dfi = trikit._cumtoincr(dfc)
    cl1 = trikit.chladder(data=dfc, trifmt="cumulative")
    cl2 = trikit.chladder(data=dfi, trifmt="incremental")
    assert cl0().equals(cl1)==cl1().equals(cl2)







# Testing _BootstrapChainLadder ==============================================]
# chladder(data, origin=None, dev=None, value=None, trifmt=None,
#              datafmt="incremental", tail=1.0, sel="all-weighted",
#              range_method=None, **kwargs):
if BOOT_CL_TEST:

    bcl0 = trikit.chladder(data=DATA, range_method="bootstrap",)


    NBR_SIMS    = 10
    prng        = RandomState(20160516)
    tri         = bcl0.tri
    latest      = bcl0.tri.latest_by_origin.sort_index()
    sel         = bcl0.sel
    tail        = bcl0.tail
    nh          = bcl0.neg_handler
    ldfs0       = bcl0.ldfs
    cldfs0      = bcl0.cldfs
    tsqr        = bcl0.trisqrd.round(0)
    ults0       = bcl0.ultimates
    ibnr0       = bcl0.reserves
    ncells      = bcl0.nbr_cells
    dof         = bcl0.dof
    sclpar      = bcl0.scale_param
    trifitc     = bcl0.tri_fit_cum
    trifiti     = bcl0.tri_fit_incr
    resid_us    = bcl0.resid_us
    resid_adj   = bcl0.resid_adj
    sdist       = bcl0.sampling_dist
    psamples    = bcl0._bs_samples(sims=NBR_SIMS, random_state=prng, parametric=True)
    npsamples   = bcl0._bs_samples(sims=NBR_SIMS, random_state=prng)
    pldfs       = bcl0._bs_ldfs(psamples)
    npldfs      = bcl0._bs_ldfs(npsamples)
    pforecasts  = bcl0._bs_forecasts(samples=psamples, ldfs=pldfs)
    npforecasts = bcl0._bs_forecasts(samples=npsamples, ldfs=npldfs)
    pperror     = bcl0._bs_process_error(pforecasts, procdist="gamma")
    npperror    = bcl0._bs_process_error(npforecasts, procdist="gamma")
    fit_assess  = bcl0.fit_assessment
    residsumm   = bcl0.residuals_detail


# Run bcl0.__call__ ==========================================================]

# __call__(sims=1000, procdist="gamma", parametric=False,
#          percentiles=[.75, .95], interpolation="linear",
#          summary=True, random_state=None)
#
# ============================================================================]

# _ldfs       = self._bs_ldfs(samples=_samples)
# _forecasts  = self._bs_forecasts(samples=_samples, ldfs=_ldfs)
# _proc_error = self._bs_process_error(forecasts=_forecasts, procdist=procdist)
# _reserves   = self._bs_reserves(_proc_error)






# Testing MackChainLadder ====================================================]

if MACK_CL_TEST:

    pass




if MCMC_CL_TEST:

    pass








# TESTING_STATE can be either "print" or "log" ===============================]
# TESTING_STATE = "print"
#
#
# if TESTING_STATE=="print:
#
#     # _IncrTriangle Tests =>
#     print("<>=<>=<>=<> `_IncrTriangle` => i1 [Actual Data] <>=<>=<>=<>")
#     print("i1.latest_diag     : {}".format(i1.latest_diag))
#     print("i1.dev_periods     : {}".format(i1.dev_periods))
#     print("i1.origin_yrs      : {}".format(i1.origin_yrs))
#     print("i1.get_origin(1989): {}".format(i1.get_origin(1989)))
#     print("i1.get_dev(4)      : {}".format(i1.get_dev(4)))
#     print("")
#
#     print("<>=<>=<>=<> `_IncrTriangle` => i2 [All NaNs]    <>=<>=<>=<>")
#     print("i2.latest_diag     : {}".format(i2.latest_diag))
#     print("i2.dev_periods     : {}".format(i2.dev_periods))
#     print("i2.origin_yrs      : {}".format(i2.origin_yrs))
#     print("i2.get_origin(1989): {}".format(i2.get_origin(1989)))
#     print("i2.get_dev(4)      : {}".format(i2.get_dev(4)))
#     print("")
#
#     print("<>=<>=<>=<> `_IncrTriangle` => i3 [0-rows/cols] <>=<>=<>=<>")
#     print("i3.latest_diag     : {}".format(i3.latest_diag))
#     print("i3.dev_periods     : {}".format(i3.dev_periods))
#     print("i3.origin_yrs      : {}".format(i3.origin_yrs))
#     print("i3.get_origin(1989): {}".format(i3.get_origin(1989)))
#     print("i3.get_dev(4)      : {}".format(i3.get_dev(4)))
#     print("")
#
#
#     # _CumTriangle Tests =>
#     print("<>=<>=<>=<> `_CumTriangle` - c1 - [Actual Data] <>=<>=<>=<>")
#     print("c1.latest_diag              : {}".format(c1.latest_diag))
#     print("c1.dev_periods              : {}".format(c1.dev_periods))
#     print("c1.origin_yrs               : {}".format(c1.origin_yrs))
#     print("c1.get_origin(1989)         : {}".format(c1.get_origin(1989)))
#     print("c1.get_dev(4)               : {}".format(c1.get_dev(4)))
#     print("c1.ldfs                      : {}".format(c1.ldfs))
#     print("c1.ldfs_avgs()               :\n{}\n".format(c1.ldfs_avgs()))
#     print("c1.ldfs_avgs(addl_avgs=8)    :\n{}\n".format(c1.ldfs_avgs(addl_avgs=8)))
#     print("c1.ldfs_avgs(addl_avgs=[6,8]):\n{}\n".format(c1.ldfs_avgs(addl_avgs=[6,8])))
#     print("")
#
#     print("<>=<>=<>=<> `_CumTriangle` - c2 - [All NaNs]    <>=<>=<>=<>")
#     print("c2.latest_diag              : {}".format(c2.latest_diag))
#     print("c2.dev_periods              : {}".format(c2.dev_periods))
#     print("c2.origin_yrs               : {}".format(c2.origin_yrs))
#     print("c2.get_origin(1989)         : {}".format(c2.get_origin(1989)))
#     print("c2.get_dev(4)               : {}".format(c2.get_dev(4)))
#     print("c2.ldfs                      : {}".format(c2.ldfs))
#     print("c2.ldfs_avgs()               :\n{}\n".format(c2.ldfs_avgs()))
#     print("c2.ldfs_avgs(addl_avgs=6)    :\n{}\n".format(c2.ldfs_avgs(addl_avgs=6)))
#     print("c2.ldfs_avgs(addl_avgs=[4,9]):\n{}\n".format(c2.ldfs_avgs(addl_avgs=[4,9])))
#     print("")
#
#     print("<>=<>=<>=<> `_CumTriangle` - c3 - [0-rows/cols] <>=<>=<>=<>")
#     print("c3.latest_diag              : {}".format(c3.latest_diag))
#     print("c3.dev_periods              : {}".format(c3.dev_periods))
#     print("c3.origin_yrs               : {}".format(c3.origin_yrs))
#     print("c3.get_origin(1989)         : {}".format(c3.get_origin(1989)))
#     print("c3.get_dev(4)               : {}".format(c3.get_dev(4)))
#     print("c3.ldfs                      : {}".format(c3.ldfs))
#     print("c3.ldfs_avgs()               :\n{}\n".format(c3.ldfs_avgs()))
#     print("c3.ldfs_avgs(addl_avgs=4)    :\n{}\n".format(c3.ldfs_avgs(addl_avgs=4)))
#     print("c3.ldfs_avgs(addl_avgs=[6,4]):\n{}\n".format(c3.ldfs_avgs(addl_avgs=[6,4])))
#     print("")
#
#
#
#
# elif TESTING)STATE=="log":
#
#     logging.basicConfig(
#             filename="U:/trikit/Logging/trikit_unittests.txt",
#             level=logging.DEBUG,
#             format='%(asctime)s - %(levelname)s - %(message)s'
#             )
#
#     # _IncrTriangle Tests =>
#     logging.debug("<>=<>=<>=<> `_IncrTriangle` => i1 [Actual Data] <>=<>=<>=<>")
#     logging.debug("i1.latest_diag     : {}".format(i1.latest_diag))
#     logging.debug("i1.dev_periods     : {}".format(i1.dev_periods))
#     logging.debug("i1.origin_yrs      : {}".format(i1.origin_yrs))
#     logging.debug("i1.get_origin(1989): {}".format(i1.get_origin(1989)))
#     logging.debug("i1.get_dev(4)      : {}".format(i1.get_dev(4)))
#     logging.debug("")
#
#     logging.debug("<>=<>=<>=<> `_IncrTriangle` => i2 [All NaNs]    <>=<>=<>=<>")
#     logging.debug("i2.latest_diag     : {}".format(i2.latest_diag))
#     logging.debug("i2.dev_periods     : {}".format(i2.dev_periods))
#     logging.debug("i2.origin_yrs      : {}".format(i2.origin_yrs))
#     logging.debug("i2.get_origin(1989): {}".format(i2.get_origin(1989)))
#     logging.debug("i2.get_dev(4)      : {}".format(i2.get_dev(4)))
#     logging.debug("")
#
#     logging.debug("<>=<>=<>=<> `_IncrTriangle` => i3 [0-rows/cols] <>=<>=<>=<>")
#     logging.debug("i3.latest_diag     : {}".format(i3.latest_diag))
#     logging.debug("i3.dev_periods     : {}".format(i3.dev_periods))
#     logging.debug("i3.origin_yrs      : {}".format(i3.origin_yrs))
#     logging.debug("i3.get_origin(1989): {}".format(i3.get_origin(1989)))
#     logging.debug("i3.get_dev(4)      : {}".format(i3.get_dev(4)))
#     logging.debug("")
#
#
#
#     # _CumTriangle Tests =>
#     logging.debug("<>=<>=<>=<> `_CumTriangle` - c1 - [Actual Data] <>=<>=<>=<>")
#     logging.debug("c1.latest_diag              : {}".format(c1.latest_diag))
#     logging.debug("c1.dev_periods              : {}".format(c1.dev_periods))
#     logging.debug("c1.origin_yrs               : {}".format(c1.origin_yrs))
#     logging.debug("c1.get_origin(1989)         : {}".format(c1.get_origin(1989)))
#     logging.debug("c1.get_dev(4)               : {}".format(c1.get_dev(4)))
#     logging.debug("c1.ldfs                      : {}".format(c1.ldfs))
#     logging.debug("c1.ldfs_avgs()               :\n{}\n".format(c1.ldfs_avgs()))
#     logging.debug("c1.ldfs_avgs(addl_avgs=8)    :\n{}\n".format(c1.ldfs_avgs(addl_avgs=8)))
#     logging.debug("c1.ldfs_avgs(addl_avgs=[6,8]):\n{}\n".format(c1.ldfs_avgs(addl_avgs=[6,8])))
#     logging.debug("")
#
#     logging.debug("<>=<>=<>=<> `_CumTriangle` - c2 - [All NaNs]    <>=<>=<>=<>")
#     logging.debug("c2.latest_diag              : {}".format(c2.latest_diag))
#     logging.debug("c2.dev_periods              : {}".format(c2.dev_periods))
#     logging.debug("c2.origin_yrs               : {}".format(c2.origin_yrs))
#     logging.debug("c2.get_origin(1989)         : {}".format(c2.get_origin(1989)))
#     logging.debug("c2.get_dev(4)               : {}".format(c2.get_dev(4)))
#     logging.debug("c2.ldfs                      : {}".format(c2.ldfs))
#     logging.debug("c2.ldfs_avgs()               :\n{}\n".format(c2.ldfs_avgs()))
#     logging.debug("c2.ldfs_avgs(addl_avgs=6)    :\n{}\n".format(c2.ldfs_avgs(addl_avgs=6)))
#     logging.debug("c2.ldfs_avgs(addl_avgs=[4,9]):\n{}\n".format(c2.ldfs_avgs(addl_avgs=[4,9])))
#     logging.debug("")
#
#     logging.debug("<>=<>=<>=<> `_CumTriangle` - c3 - [0-rows/cols] <>=<>=<>=<>")
#     logging.debug("c3.latest_diag              : {}".format(c3.latest_diag))
#     logging.debug("c3.dev_periods              : {}".format(c3.dev_periods))
#     logging.debug("c3.origin_yrs               : {}".format(c3.origin_yrs))
#     logging.debug("c3.get_origin(1989)         : {}".format(c3.get_origin(1989)))
#     logging.debug("c3.get_dev(4)               : {}".format(c3.get_dev(4)))
#     logging.debug("c3.ldfs                      : {}".format(c3.ldfs))
#     logging.debug("c3.ldfs_avgs()               :\n{}\n".format(c3.ldfs_avgs()))
#     logging.debug("c3.ldfs_avgs(addl_avgs=4)    :\n{}\n".format(c3.ldfs_avgs(addl_avgs=4)))
#     logging.debug("c3.ldfs_avgs(addl_avgs=[6,4]):\n{}\n".format(c3.ldfs_avgs(addl_avgs=[6,4])))
#     logging.debug("")



"""
This module contains the definitions of both the ``_IncrTriangle`` and
``_CumTriangle`` classes. Users should avoid instantiating ``_IncrTriangle``
or ``_CumTriangle`` instances directly; rather the dataset and triangle
arguments should be passed to ``totri``, which will return either an
instance of ``_CumTriangle`` or ``_IncrTriangle``, depending on the argument
specified for ``type_``.

##### TODO #####
[0] Refactor a2a_avgs
[1] Integrate a2aind into a2a_avgs
[2] Make a2a_avgs a function rather than a property.
"""
import itertools
import numpy as np
import pandas as pd


from .chainladder import _BaseChainLadder
#from .chainladder import bootstrap._BootstrapChainLadder
from .chainladder.bootstrap import _BootstrapChainLadder
from .chainladder.mack import _MackChainLadder
# from chainladder.bootstrap import _BootstrapChainLadder


class _IncrTriangle(pd.DataFrame):
    """
    ``_IncrTriangle`` specification. Note that this class isn't part of
    trikit's public interface, as is used  primarily as a base class for
    ``_CumTriangle`` object definitions. As such, users shouldn't
    instantiate ``_IncrTriangle`` instances directly. To obtain an
    incremental triangle object, use ``trikit.totri``.
    """
    def __init__(self, data, origin=None, dev=None, value=None):
        """
        Attempts to convert ``data`` to an incremental triangle instance.
        Checks for ``origin``, ``dev`` and ``value`` arguments for fieldnames
        corresponding to origin year, development period and loss amount
        respectively. If ``origin``, ``dev`` and ``value`` are unspecified,
        fieldnames are assumed to be "origin", "dev" and "value".

        Parameters
        ----------
        data: pd.DataFrame
            The dataset to be coerced into a ``_IncrTriangle`` instance.
            ``data`` must be tabular loss data with at minimum columns
            representing the origin/acident year, the development
            period and the actual loss amount, given by ``origin``, ``dev``
            and ``value`` arguments.

        origin: str
            The fieldname in ``data`` representing the origin year.
            Defaults to None.

        dev: str
            The fieldname in ``data`` representing the development period.
            Defaults to None.

        value: str
            The fieldname in ``data`` representing loss amounts.
            Defaults to None.

        Returns
        -------
        trikit.triangle._IncrTriangle
        """
        try:
            if all(i is None for i in (origin, dev, value)):
                origin, dev, value = "origin", "dev", "value"
            data2 = data[[origin, dev, value]]
            data2 = data2.groupby([origin, dev], as_index=False).sum()
            tri = data2.pivot(index=origin, columns=dev).rename_axis(None)
            tri.columns = tri.columns.droplevel(0)

        except KeyError:
            print("One or more fields not present in data.")

        # Force all triangle cells to be of type np.float_.
        for i in tri:
            tri[i] = tri[i].astype(np.float_)

        tri.columns.name = None

        super().__init__(tri)

        self.origin = origin
        self.value = value
        self.dev = dev

        # Properties.
        self._latest_by_origin = None
        self._latest_by_devp = None
        self._nbr_cells = None
        self._maturity = None
        self._triind = None
        self._devp = None
        self._latest = None
        self._origins = None
        self._rlvi = None
        self._clvi = None
        self._dof = None



    @property
    def nbr_cells(self):
        """
        Return the number of non-NaN cells.

        Returns
        -------
        int
        """
        if self._nbr_cells is None:
            self._nbr_cells = self.count().sum()
        return(self._nbr_cells)


    @property
    def dof(self):
        """
        Return the degress of freedom.

        Returns
        -------
        int
        """
        if self._dof is None:
            self._dof = self.nbr_cells - (self.columns.size-1) + self.index.size
        return(self._dof)


    @property
    def triind(self):
        """
        Table indicating forecast cells with 1, actual data with 0.

        Returns
        -------
        pd.DataFrame
        """
        if self._triind is None:
            self._triind = self.applymap(lambda x: 1 if np.isnan(x) else 0)
        return(self._triind)


    @property
    def rlvi(self):
        """
        Determine the last valid index by row/origin period.

        Returns
        -------
        pd.DataFrame
        """
        if self._rlvi is None:
            self._rlvi = pd.DataFrame({
                "dev":self.apply(
                    lambda x: x.last_valid_index(), axis=1).values
                },index=self.index)
            self._rlvi["col_offset"] = \
                self._rlvi["dev"].map(lambda x: self.columns.get_loc(x))
        return(self._rlvi)


    @property
    def clvi(self):
        """
        Determine the last valid index by development period.

        Returns
        -------
        pd.DataFrame
        """
        if self._clvi is None:
            self._clvi = pd.DataFrame({
                "origin":self.apply(lambda x: x.last_valid_index(), axis=0).values
                },index=self.columns)
            self._clvi["row_offset"] = \
                self._clvi["origin"].map(lambda x: self.index.get_loc(x))
        return(self._clvi)


    @property
    def latest(self):
        """
        Return the values on the triangle's latest diagonal. Loss amounts
        are given, along with the associated origin year and development
        period. The latest loss amount by origin year alone can be  obtained
        by calling ``self.latest_by_origin``, or by development period by
        calling by ``self.latest_by_devp``.

        Returns
        -------
        pd.DataFrame
        """
        if self._latest is None:
            lindx = self.apply(lambda dev_: dev_.last_valid_index(), axis=1)
            self._latest = pd.DataFrame(
                {"latest":self.lookup(lindx.index, lindx.values),
                 "origin":lindx.index, "dev":lindx.values})
        return(self._latest[["origin", "dev", "latest"]].sort_index())


    @property
    def latest_by_origin(self):
        """
        Return the latest loss amounts by origin year.

        Returns
        -------
        pd.Series
        """
        if self._latest_by_origin is None:
            self._latest_by_origin = pd.Series(
                data=self.latest["latest"].values, index=self.latest["origin"].values,
                name="latest_by_origin")
        return(self._latest_by_origin.sort_index())


    @property
    def latest_by_devp(self):
        """
        Return the latest loss amounts by development period.

        Returns
        -------
        pd.Series
        """
        if self._latest_by_devp is None:
            self._latest_by_devp = pd.Series(
                data=self.latest["latest"].values, index=self.latest["dev"].values,
                name="latest_by_devp")
        return(self._latest_by_devp.sort_index())


    @property
    def devp(self):
        """
        Return triangle's development periods.

        Returns
        -------
        pd.Series
        """
        if self._devp is None:
            self._devp = pd.Series(self.columns,name="dev")
        return(self._devp.sort_index())


    @property
    def origins(self):
        """
        Return triangle's origin periods.

        Returns
        -------
        pd.Series
        """
        if self._origins is None:
            self._origins = pd.Series(self.index, name="origin")
        return(self._origins.sort_index())


    @property
    def maturity(self):
        """
        Return the maturity for each origin period.

        Returns
        -------
        ps.Series
        """
        if self._maturity is None:
            dfind, matlist  = (1 - self.triind), list()
            for i in range(dfind.index.size):
                lossyear = dfind.index[i]
                maxindex = dfind.loc[lossyear].to_numpy().nonzero()[0].max()
                itermatur = dfind.columns[maxindex]
                matlist.append(itermatur)
            self._maturity = pd.Series(data=matlist, index=self.index, name="maturity")
        return(self._maturity.sort_index())


    def as_tbl(self):
        """
        Transform triangle instance into a tabular representation.

        Returns
        -------
        pd.DataFrame
        """
        tri_ = self.reset_index(drop=False).rename({"index":"origin"}, axis=1)
        df_ = pd.melt(tri_, id_vars=[self.origin], var_name=self.dev, value_name=self.value)
        df_ = df_[~np.isnan(df_[self.value])]
        df_ = df_.astype({self.origin:np.int_, self.dev:np.int_, self.value:np.float_})
        return(df_.sort_values(by=[self.origin, self.dev]).reset_index(drop=True))


    def as_cum(self):
        """
        Transform incremental triangle instance into a cumulative
        representation. Note that returned object will be a DataFrame,
        not an instance of ``triangle._CumTriangle``.

        Returns
        -------
        pd.DataFrame
        """
        return(self.cumsum(axis=1))


    def as_incr(self):
        """
        Transform ``triangle._IncrTriangle`` instance to pd.DataFrame.

        Returns
        -------
        pd.DataFrame
        """
        return(pd.DataFrame(self))


    # def __str__(self):
    #     # Controls when print(self) is called
    #     formats_ = {dev_:"{:.0f}".format for dev_ in self.columns}
    #     return(self.to_string(formatters=formats_))


    def __repr__(self):
        """
        Controls when object is referenced from interpreter.
        """
        formats_ = {dev_:"{:.0f}".format for dev_ in self.columns}
        return(self.to_string(formatters=formats_))





class _CumTriangle(_IncrTriangle):
    """
    Cumulative triangle class definition.
    """
    def __init__(self, data, origin=None, dev=None, value=None):
        """
        Attempts to represent ``data`` as a cumulative triangle instance.
        ``origin``, ``dev`` and ``value`` arguments represent fieldnames
        corresponding to loss year, development period and loss amount
        respectively. If ``origin``, ``dev`` and ``value`` are unspecified,
        fieldnames are assumed to be "origin", "dev" and "value".

        Parameters
        ----------
        origin: str
            The fieldname in ``data`` representing the origin year.
            Defaults to None.

        dev: str
            The fieldname in ``data`` representing the development period.
            Defaults to None.

        value: str
            The fieldname in ``data`` representing loss amounts.
            Defaults to None.

        Returns
        -------
        trikit.triangle._CumTriangle
        """
        if all(i is None for i in (origin, dev, value)):
            origin, dev, value = "origin", "dev", "value"
        data2 = data[[origin, dev, value]]
        data2 = data2.groupby([origin, dev], as_index=False).sum()
        data2 = data2.sort_values(by=[origin, dev])
        data2["cumval"] = data2.groupby([origin], as_index=False)[value].cumsum()
        data2 = data2.drop(value, axis=1)
        data2 = data2.rename(columns={"cumval":value})
        super().__init__(data=data2, origin=origin, dev=dev, value=value)

        # Properties
        self._a2a_avgs = None
        self._a2aind = None
        self._a2a = None


    @staticmethod
    def _geometric(vals, weights=None):
        """
        Compute the geometric average of the elements of ``vals``.

        Parameters
        ----------
        vals: np.ndarray
            An array of values, typically representing link ratios from a
            single development period.

        weights: np.ndarray
            Weights to assign specific values in the average computation.
            If None, each value is assigned equal weight.

        Returns
        -------
        float
        """
        if len(vals)==0:
            avg_ = None
        else:
            avg_ = np.prod(vals) ** (1 / len(vals))
        return(avg_)


    @staticmethod
    def _simple(vals, weights=None):
        """
        Compute the simple average of elements of ``vals``.

        Parameters
        ----------
        vals: np.ndarray
            An array of values, typically representing link ratios from a
            single development period.

        weights: np.ndarray
            Weights to assign specific values in the average computation.
            If None, each value is assigned equal weight.

        Returns
        -------
        float
        """
        if len(vals) < 1:
            avg_ = None
        else:
            avg_ = sum(vals) / len(vals)
        return(avg_)


    @staticmethod
    def _medial(vals, weights=None):
        """
        Compute the medial average of elements in ``vals``. Medial average
        eliminates the min and max values, then returns the arithmetic
        average of the remaining items.

        Parameters
        ----------
        vals: np.ndarray
            An array of values, typically representing link ratios from a
            single development period.

        weights: np.ndarray
            Weights to assign specific values in the average computation.
            If None, each value is assigned equal weight.

        Returns
        -------
        float
        """
        vals = list(vals)
        if len(vals)==0:
            avg_ = None
        elif len(vals)==1:
            avg_ = vals[0]
        elif len(vals)==2:
            avg_ = sum(vals) / len(vals)
        else:
            keep_ = sorted(vals)[1:-1]
            avg_ = sum(keep_) / len(keep_)
        return(avg_)


    @property
    def a2a(self):
        """
        Compute adjacent proportions, a.k.a. link ratios.

        Returns
        -------
        pd.DataFrame
        """
        if self._a2a is None:
            self._a2a = self.shift(periods=-1, axis=1) / self
            self._a2a = self._a2a.dropna(axis=1, how="all").dropna(axis=0, how="all")
        return(self._a2a.sort_index())


    @property
    def a2aind(self):
        """
        Determine which cells should be included and which to exclude
        when computing age-to-age averages.

        Returns
        -------
        pd.DataFrame
        """
        if self._a2aind is None:
            self._a2aind = self.a2a.applymap(lambda v: 0 if np.isnan(v) else 1)
        return(self._a2aind)


    @a2aind.setter
    def a2aind(self, update_spec):
        """
        Update ``self.a2aind`` in order to down-weight ldfs in Chain Ladder
        calculation.

        Parameters
        ----------
        update_spec: tuple
            3-tuple consisting of ``(index, column, value)``, representing
            the intersection point of the ``self.a2a`` target cell, and the
            value used to update it.

        Examples
        --------
        Load raa sample dataset, and remove a highly-leveraged age-to-age
        factor from influencing the ldf calculation.

        >>> import trikit
        >>> raa = trikit.load(dataset="raa")
        >>> tri = trikit.totri(data=raa)
        >>> tri.a2a.iloc[:, :1]
                      1
        1981   1.649840
        1982  40.424528
        1983   2.636950
        1984   2.043324
        1985   8.759158
        1986   4.259749
        1987   7.217235
        1988   5.142117
        1989   1.721992

        To remove the link ratio at origin year 1982 and development
        period 1, run the following:

        >>> tri.a2aind = (1982, 1, 0)
        >>> tri.a2aind
              1  2  3  4  5  6  7  8  9
        1981  1  1  1  1  1  1  1  1  1
        1982  0  1  1  1  1  1  1  1  0
        1983  1  1  1  1  1  1  1  0  0
        1984  1  1  1  1  1  1  0  0  0
        1985  1  1  1  1  1  0  0  0  0
        1986  1  1  1  1  0  0  0  0  0
        1987  1  1  1  0  0  0  0  0  0
        1988  1  1  0  0  0  0  0  0  0
        1989  1  0  0  0  0  0  0  0  0

        Notice that the value at (1982, 1) is 0. To change it back
        to 1, simply run:

        >>> tri.a2aind = (1982, 1, 1)
        >>> tri.a2aind
              1  2  3  4  5  6  7  8  9
        1981  1  1  1  1  1  1  1  1  1
        1982  1  1  1  1  1  1  1  1  0
        1983  1  1  1  1  1  1  1  0  0
        1984  1  1  1  1  1  1  0  0  0
        1985  1  1  1  1  1  0  0  0  0
        1986  1  1  1  1  0  0  0  0  0
        1987  1  1  1  0  0  0  0  0  0
        1988  1  1  0  0  0  0  0  0  0
        1989  1  0  0  0  0  0  0  0  0

        Note also that ``self.a2aind`` may be updated using DataFrame
        methods directly:

        >>> tri.a2aind.at[1982, 1] = 0
        >>> tri.a2aind
              1  2  3  4  5  6  7  8  9
        1981  1  1  1  1  1  1  1  1  1
        1982  0  1  1  1  1  1  1  1  0
        1983  1  1  1  1  1  1  1  0  0
        1984  1  1  1  1  1  1  0  0  0
        1985  1  1  1  1  1  0  0  0  0
        1986  1  1  1  1  0  0  0  0  0
        1987  1  1  1  0  0  0  0  0  0
        1988  1  1  0  0  0  0  0  0  0
        1989  1  0  0  0  0  0  0  0  0
        """
        indx, column, value = update_spec
        self._a2aind.at[indx, column] = value


    @property
    def a2a_avgs(self):
        """
        Compute age-to-age factors based on ``self.a2a`` table of adjacent
        proportions. Averages computed include "simple", "geometric", "medial"
        and "weighted".

        Returns
        -------
        pd.DataFrame
        """
        if self._a2a_avgs is None:

            _nbr_periods = list(range(1, self.a2a.shape[0])) + [0]
            indxstrs = list()

            # Create lookup table for average functions.
            avgfuncs = {
                'simple'   :self._simple,
                'geometric':self._geometric,
                'medial'   :self._medial,
                'weighted' :None
                }

            # Remove `0` entry, and add as last element of list.
            ldf_avg_lst = list(itertools.product(avgfuncs.keys(), _nbr_periods))

            indxstrs = [
                "all-" + str(i[0]) if i[1]==0 else "{}-{}".format(i[0], i[1])
                    for i in ldf_avg_lst
                ]

            # for i in ldf_avg_lst:
            #     iteravg, iterdur = i[0], i[1]
            #     iterstr = "all-" + str(iteravg) if iterdur==0 \
            #               else str(iterdur) + "-" + str(iteravg)
            #     indxstrs.append(iterstr)

            indx = sorted(ldf_avg_lst, key=lambda x: x[1])
            self._a2a_avgs = pd.DataFrame(index=indxstrs, columns=self.a2a.columns)

            for a in enumerate(ldf_avg_lst):
                duration, avgtype, indxpos = a[1][1], a[1][0], a[0]
                indxstr, iterfunc = indxstrs[indxpos], avgfuncs[avgtype]
                for col in range(self.a2a.shape[1]):
                    itercol, colstr = self.a2a.iloc[:, col], self.a2a.columns[col]

                    if avgtype=='weighted':
                        t_ic_1, t_ic_2 = self.iloc[:, col], self.iloc[:, (col + 1)]
                        # Find first NaN value in t_ic_2.
                        first_nan_year = t_ic_2.index[t_ic_2.count():][0]
                        first_nan_indx = t_ic_2.index.searchsorted(first_nan_year)
                        final_cell_indx = first_nan_indx
                        if duration==0:
                            first_cell_indx = 0
                        else:
                            first_cell_indx = (final_cell_indx-duration) if \
                                              (final_cell_indx-duration)>=0 else 0

                        # Divide sum of t_ic_2 by t_ic_1.
                        ic_2 = t_ic_2[first_cell_indx:final_cell_indx]
                        ic_1 = t_ic_1[first_cell_indx:final_cell_indx]
                        sum_ic_2 = t_ic_2[first_cell_indx:final_cell_indx].sum()
                        sum_ic_1 = t_ic_1[first_cell_indx:final_cell_indx].sum()

                        try:
                            iteravg = (sum_ic_2/sum_ic_1)
                        except ZeroDivisionError:
                            iteravg = np.NaN

                    else: # avgtype in ('simple', 'geometric', 'medial')
                        # find index of first row with NaN
                        if any(itercol.map(lambda x: np.isnan(x))):
                            first_nan_year = \
                                itercol.index[itercol.apply(lambda x: np.isnan(x))][0]
                            first_nan_indx = \
                                itercol.index.searchsorted(first_nan_year)
                            final_cell_indx = first_nan_indx
                            if duration==0:
                                first_cell_indx = 0
                            else:
                                first_cell_indx = (final_cell_indx-duration) if \
                                                  (final_cell_indx-duration)>=0 else 0

                        else: # itercol has 0 NaN's
                            final_cell_indx = len(itercol)
                            first_cell_indx = 0 if duration==0 else (final_cell_indx-duration)
                        try:
                            iteravg = iterfunc(itercol[first_cell_indx:final_cell_indx])

                        except ZeroDivisionError:
                            iteravg = np.NaN

                    self._a2a_avgs.loc[indxstr, colstr] = iteravg

        return(self._a2a_avgs)


    def plot(self, color="#334488", axes_style="darkgrid", context="notebook",
             col_wrap=5, **kwargs):
        """
        Visualize triangle development patterns.

        Parameters
        ----------
        facets: bool
            If True, loss development plots for each origin year are faceted,
            otherwise development patterns are plotted together on a single set
            of axes.

        axes_style: str
            Aesthetic style of plots. Defaults to "darkgrid". Other options
            include: {whitegrid, dark, white, ticks}.

        context: str
            Set the plotting context parameters. According to the seaborn
            documentation, This affects things like the size of the labels,
            lines, and other elements of the plot, but not the overall style.
            Defaults to ``"notebook"``. Additional options include
            {paper, talk, poster}.

        palette: str
            Selected matplotlib color map. For additional options, visit:
            https://matplotlib.org/tutorials/colors/colormaps.html.

        kwargs: dict
            Additional plot styling options.

        Returns
        -------
        matplotlib.pyplot.plot
        """
        import matplotlib as mpl
        import matplotlib.pyplot as plt
        import seaborn as sns

        sns.set_context(context)
        data = self.as_tbl()

        with sns.axes_style(axes_style):

            pltkwargs = dict(
                marker="o", markersize=7, alpha=1, markeredgecolor="#000000",
                markeredgewidth=.50, linestyle="--", linewidth=.75,
                fillstyle="full", color=color,
                )

            if kwargs:
                pltkwargs.update(kwargs)

            titlestr_ = "Actual Loss Development by Origin"

            g = sns.FacetGrid(
                data, col="origin", col_wrap=col_wrap, margin_titles=False,
                despine=True, sharex=True, sharey=True,
                )

            g.map(plt.plot, "dev", "value", **pltkwargs)
            g.set_axis_labels("", "")
            g.set(xticks=data.dev.unique().tolist())
            g.set_titles("{col_name}", size=9)
            g.set_xticklabels(data.dev.unique().tolist(), size=8)

            # Change ticklabel font size and place legend on each facet.
            for i, _ in enumerate(g.axes):
                ax_ = g.axes[i]
                ylabelss_ = [i.get_text() for i in list(ax_.get_yticklabels())]
                ylabelsn_ = [float(i.replace(u"\u2212", "-")) for i in ylabelss_]
                ylabelsn_ = [i for i in ylabelsn_ if i>=0]
                ylabels_ = ["{:,.0f}".format(i) for i in ylabelsn_]
                ax_.set_yticklabels(ylabels_, size=8)

                # Draw border around each facet.
                for _, spine_ in ax_.spines.items():
                    spine_.set_visible(True)
                    spine_.set_color("#000000")
                    spine_.set_linewidth(.50)

        # Adjust facets downward and and left-align figure title.
        plt.subplots_adjust(top=0.9)
        g.fig.suptitle(
            titlestr_, x=0.065, y=.975, fontsize=11, color="#404040", ha="left"
            )


    def as_incr(self):
        """
        Convert cumulative triangle instance into an incremental
        representation. Note that returned object will not be of type
        ``triangle._IncrTriangle``, but will instead be a DataFrame.

        Returns
        -------
        ps.DataFrame
        """
        tri_ = self.diff(axis=1)
        tri_.iloc[:, 0] = self.iloc[:, 0]
        return(tri_)


    def as_cum(self):
        """
        Transform ``triangle._CumTriangle`` instance to pd.DataFrame.

        Returns
        -------
        pd.DataFrame
        """
        return(pd.DataFrame(self))


    def _chladder(self, sel="all-weighted", tail=1.0, range_method=None, **kwargs):
        """
        trikit's chain ladder implementation.
        """
        if range_method is None:
            cl_ = _BaseChainLadder(self).run(sel="all-weighted", tail=1.0)
        elif range_method.lower().strip().startswith("boot"):
            cl_ = _BootstrapChainLadder(self).run(sel="all-weighted", **kwargs)
        return(cl_)



